#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007 Donald N. Allingham
# Copyright (C) 2007-2012 Brian G. Matherly
# Copyright (C) 2009      Nick Hall
# Copyright (C) 2009      Benny Malengier
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2011      Tim G L Lyons
# Copyright (C) 2012      Mathieu MD
# Copyright (C) 2013-2016 Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

""" Complete Individual Report """

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------
import os
from collections import defaultdict

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import EventRoleType, EventType, Person, NoteType
from gramps.gen.plug.docgen import (
    IndexMark,
    FontStyle,
    ParagraphStyle,
    TableStyle,
    TableCellStyle,
    FONT_SANS_SERIF,
    INDEX_TYPE_TOC,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_RIGHT,
)
from gramps.gen.display.place import displayer as _pd
from gramps.gen.plug.menu import (
    BooleanOption,
    FilterOption,
    PersonOption,
    BooleanListOption,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import Bibliography
from gramps.gen.plug.report import endnotes as Endnotes
from gramps.gen.plug.report import stdoptions
from gramps.gen.utils.file import media_path_full
from gramps.gen.utils.lds import TEMPLES
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.errors import ReportError

# ------------------------------------------------------------------------
#
# Global variables (ones used in both classes here, that is)
#
# ------------------------------------------------------------------------


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


CUSTOM = _T_("Custom")

# Construct section list and type to group map
SECTION_LIST = []
TYPE2GROUP = {}
for event_group, type_list in EventType().get_menu_standard_xml():
    SECTION_LIST.append(event_group)
    for event_type in type_list:
        TYPE2GROUP[event_type] = event_group
SECTION_LIST.append(CUSTOM)
TYPE2GROUP[EventType.CUSTOM] = CUSTOM
TYPE2GROUP[EventType.UNKNOWN] = _T_("Unknown")


# ------------------------------------------------------------------------
#
# IndivCompleteReport
#
# ------------------------------------------------------------------------
class IndivCompleteReport(Report):
    """the Complete Individual Report"""

    def __init__(self, database, options, user):
        """
        Create the IndivCompleteReport object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        filter    - Filter to be applied to the people of the database.
                    The option class carries its number, and the function
                    returning the list of filters.
        cites     - Whether or not to include source information.
        sort      - Whether or not to sort events into chronological order.
        grampsid  - Whether or not to include any Gramps IDs
        images    - Whether or not to include images.
        sections  - Which event groups should be given separate sections.
        name_format   - Preferred format to display names
        incl_private  - Whether to include private data
        incl_attrs    - Whether to include attributes
        incl_census   - Whether to include census events
        incl_notes    - Whether to include person and family notes
        incl_tags     - Whether to include tags
        incl_relname  - Whether to include relationship to home person
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """

        Report.__init__(self, database, options, user)
        self._user = user
        menu = options.menu

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)
        self._db = self.database

        self.use_pagebreak = menu.get_option_by_name("pageben").get_value()

        self.sort = menu.get_option_by_name("sort").get_value()

        self.name_is_title = menu.get_option_by_name("name_title").get_value()

        self.use_attrs = menu.get_option_by_name("incl_attrs").get_value()
        self.use_census = menu.get_option_by_name("incl_census").get_value()
        self.use_gramps_id = menu.get_option_by_name("inc_id").get_value()
        self.use_images = menu.get_option_by_name("images").get_value()
        self.use_notes = menu.get_option_by_name("incl_notes").get_value()
        self.use_srcs = menu.get_option_by_name("cites").get_value()
        self.use_src_notes = menu.get_option_by_name("incsrcnotes").get_value()
        self.use_tags = menu.get_option_by_name("incl_tags").get_value()

        filter_option = options.menu.get_option_by_name("filter")
        self.filter = filter_option.get_filter()

        self.section_list = menu.get_option_by_name("sections").get_selected()

        stdoptions.run_name_format_option(self, menu)

        self.place_format = menu.get_option_by_name("place_format").get_value()

        self.home_person = self._db.get_default_person()  # might be None
        self.increlname = menu.get_option_by_name("incl_relname").get_value()
        if self.increlname and self.home_person:
            from gramps.gen.relationship import get_relationship_calculator

            self.rel_calc = get_relationship_calculator(
                reinit=True, clocale=self._locale
            )

        self.bibli = None
        self.family_notes_list = []
        self.names_notes_list = []
        self.mime0 = None
        self.person = None

    def write_fact(self, event_ref, event, show_type=True):
        """
        Writes a single event.
        """
        role = event_ref.get_role()
        description = event.get_description()

        date = self._get_date(event.get_date_object())
        place_name = ""
        place_endnote = ""
        place_handle = event.get_place_handle()
        if place_handle:
            place = self._db.get_place_from_handle(place_handle)
            place_name = _pd.display_event(self._db, event, self.place_format)
            place_endnote = self._cite_endnote(place)
        # make sure it's translated, so it can be used below, in "combine"
        ignore = _("%(str1)s in %(str2)s. ") % {"str1": "", "str2": ""}
        date_place = self.combine("%(str1)s in %(str2)s. ", "%s. ", date, place_name)

        if show_type:
            # Groups with more than one type
            column_1 = self._(self._get_type(event.get_type()))
            if role not in (EventRoleType.PRIMARY, EventRoleType.FAMILY):
                column_1 = column_1 + " (" + self._(role.xml_str()) + ")"
            # Translators: needed for Arabic, ignore otherwise
            # make sure it's translated, so it can be used below, in "combine"
            ignore = _("%(str1)s, %(str2)s") % {"str1": "", "str2": ""}
            column_2 = self.combine("%(str1)s, %(str2)s", "%s", description, date_place)
        else:
            # Groups with a single type (remove event type from first column)
            column_1 = date
            # Translators: needed for Arabic, ignore otherwise
            # make sure it's translated, so it can be used below, in "combine"
            ignore = _("%(str1)s, %(str2)s") % {"str1": "", "str2": ""}
            column_2 = self.combine("%(str1)s, %(str2)s", "%s", description, place_name)

        event_ref_endnotes = self._cite_endnote(event_ref, prior=place_endnote)
        endnotes = self._cite_endnote(event, prior=event_ref_endnotes)

        self.doc.start_row()
        self.write_cell(column_1)
        self.doc.start_cell("IDS-NormalCell")
        self.doc.start_paragraph("IDS-Normal")
        self.doc.write_text(column_2)
        if endnotes:
            self.doc.start_superscript()
            self.doc.write_text(endnotes)
            self.doc.end_superscript()
        self.doc.end_paragraph()

        self.do_attributes(event.get_attribute_list() + event_ref.get_attribute_list())

        for notehandle in event.get_note_list():
            note = self._db.get_note_from_handle(notehandle)
            text = note.get_styledtext()
            if self.use_gramps_id:
                text = text + " [%s]" % note.get_gramps_id()
            note_format = note.get_format()
            self.doc.write_styled_note(
                text,
                note_format,
                "IDS-Normal",
                contains_html=(note.get_type() == NoteType.HTML_CODE),
            )

        self.doc.end_cell()
        self.doc.end_row()

    def write_p_entry(self, label, parent_name, rel_type, pmark=None):
        """write parent entry"""
        self.doc.start_row()
        self.write_cell(label)
        if parent_name:
            # Translators: e.g. (a stepfather): John Smith, relationship: Step
            text = self._("%(parent-name)s, relationship: %(rel-type)s") % {
                "parent-name": parent_name,
                "rel-type": self._(rel_type),
            }
            self.write_cell(text, mark=pmark)
        else:
            self.write_cell("")
        self.doc.end_row()

    def write_note(self):
        """write a note"""
        notelist = self.person.get_note_list()[
            :
        ]  # we don't want to modify cached original
        notelist += self.family_notes_list
        if self.names_notes_list:
            for note_handle in self.names_notes_list:
                if note_handle not in notelist:
                    notelist += [note_handle]
        if not notelist or not self.use_notes:
            return
        self.doc.start_table("note", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Notes"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()

        for notehandle in notelist:
            note = self._db.get_note_from_handle(notehandle)
            text = note.get_styledtext()
            if self.use_gramps_id:
                text = text + " [%s]" % note.get_gramps_id()
            note_format = note.get_format()
            self.doc.start_row()
            self.doc.start_cell("IDS-NormalCell", 2)
            self.doc.write_styled_note(
                text,
                note_format,
                "IDS-Normal",
                contains_html=(note.get_type() == NoteType.HTML_CODE),
            )

            self.doc.end_cell()
            self.doc.end_row()

        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_alt_parents(self):
        """write any alternate parents"""

        family_handle_list = self.person.get_parent_family_handle_list()
        if len(family_handle_list) < 2:
            return

        self.doc.start_table("altparents", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Alternate Parents"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()

        for family_handle in family_handle_list:
            if family_handle == self.person.get_main_parents_family_handle():
                continue

            family = self._db.get_family_from_handle(family_handle)

            # Get the mother and father relationships
            frel = ""
            mrel = ""
            child_handle = self.person.get_handle()
            child_ref_list = family.get_child_ref_list()
            for child_ref in child_ref_list:
                if child_ref.ref == child_handle:
                    frel = str(child_ref.get_father_relation())
                    mrel = str(child_ref.get_mother_relation())

            father_handle = family.get_father_handle()
            if father_handle:
                father = self._db.get_person_from_handle(father_handle)
                fname = self._name_display.display(father)
                mark = utils.get_person_mark(self._db, father)
                self.write_p_entry(self._("Father"), fname, frel, mark)
            else:
                self.write_p_entry(self._("Father"), "", "")

            mother_handle = family.get_mother_handle()
            if mother_handle:
                mother = self._db.get_person_from_handle(mother_handle)
                mname = self._name_display.display(mother)
                mark = utils.get_person_mark(self._db, mother)
                self.write_p_entry(self._("Mother"), mname, mrel, mark)
            else:
                self.write_p_entry(self._("Mother"), "", "")

        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def get_name(self, person):
        """prepare the name to display"""
        name = self._name_display.display(person)
        if self.use_gramps_id:
            return "%(name)s [%(gid)s]" % {"name": name, "gid": person.get_gramps_id()}
        else:
            return name

    def write_home_relationship(self):
        """write the person's relationship to the home/default person"""

        if not (
            self.increlname and self.home_person and self.home_person != self.person
        ):
            return
        home_relationship = self.rel_calc.get_one_relationship(
            self._db, self.home_person, self.person, olocale=self._locale
        )
        if not home_relationship:
            return

        self.doc.start_table("homerelation", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(
            self._("Relationship to home person"), style="IDS-SectionTitle"
        )
        self.doc.end_cell()
        self.doc.end_row()
        self.doc.start_row()
        self.write_cell(home_relationship, span=2)
        self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_alt_names(self):
        """write any alternate names of the person"""

        alt_names = self.person.get_alternate_names()
        if len(alt_names) < 1:
            return

        self.doc.start_table("altnames", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Alternate Names"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()

        for name in alt_names:
            self.names_notes_list += name.get_note_list()
            name_type = self._(self._get_type(name.get_type()))
            self.doc.start_row()
            self.write_cell(name_type)
            text = self._name_display.display_name(name)
            endnotes = self._cite_endnote(name)
            self.write_cell(text, endnotes)
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_addresses(self):
        """write any addresses of the person"""

        alist = self.person.get_address_list()

        if len(alist) == 0:
            return

        self.doc.start_table("addresses", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Addresses"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()

        for addr in alist:
            text = utils.get_address_str(addr)
            date = self._get_date(addr.get_date_object())
            endnotes = self._cite_endnote(addr)
            self.doc.start_row()
            self.write_cell(date)
            self.write_cell(text, endnotes)
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_associations(self):
        """write any associations of the person"""

        if len(self.person.get_person_ref_list()) < 1:
            return

        self.doc.start_table("associations", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Associations"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()
        for person_ref in self.person.get_person_ref_list():
            endnotes = self._cite_endnote(person_ref)
            relationship = person_ref.get_relation()
            associate = self._db.get_person_from_handle(person_ref.ref)
            associate_name = self._name_display.display(associate)
            self.doc.start_row()
            self.write_cell(self._(relationship))
            self.write_cell(associate_name, endnotes)
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_attributes(self):
        """write any attributes of the person"""

        attr_list = self.person.get_attribute_list()

        if len(attr_list) == 0 or not self.use_attrs:
            return

        self.doc.start_table("attributes", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Attributes"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()

        for attr in attr_list:
            attr_type = attr.get_type().type2base()
            self.doc.start_row()
            self.write_cell(self._(attr_type))
            text = attr.get_value()
            endnotes = self._cite_endnote(attr)
            self.write_cell(text, endnotes)
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_LDS_ordinances(self):
        """write any LDS ordinances of the person"""

        ord_list = self.person.get_lds_ord_list()

        if len(ord_list) == 0:
            return

        self.doc.start_table("ordinances", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("LDS Ordinance"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()
        self.doc.end_table()

        self.doc.start_table("ordinances3", "IDS-OrdinanceTable")
        self.doc.start_row()
        self.write_cell(self._("Type"), style="IDS-TableSection")
        self.write_cell(self._("Date"), style="IDS-TableSection")
        self.write_cell(self._("Status"), style="IDS-TableSection")
        self.write_cell(self._("Temple"), style="IDS-TableSection")
        self.write_cell(self._("Place"), style="IDS-TableSection")
        self.doc.end_row()

        for lds_ord in ord_list:
            otype = self._(lds_ord.type2str())
            date = self._get_date(lds_ord.get_date_object())
            status = self._(lds_ord.status2str())
            temple = TEMPLES.name(lds_ord.get_temple())
            place_name = ""
            place_endnote = ""
            place_handle = lds_ord.get_place_handle()
            if place_handle:
                place = self._db.get_place_from_handle(place_handle)
                place_name = _pd.display_event(self._db, lds_ord, self.place_format)
                place_endnote = self._cite_endnote(place)
            endnotes = self._cite_endnote(lds_ord, prior=place_endnote)
            self.doc.start_row()
            self.write_cell(otype, endnotes)
            self.write_cell(date)
            self.write_cell(status)
            self.write_cell(temple)
            self.write_cell(place_name)
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_tags(self):
        """write any tags the person has"""

        thlist = self.person.get_tag_list()
        if len(thlist) == 0 or not self.use_tags:
            return
        tags = []

        self.doc.start_table("tags", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Tags"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()
        for tag_handle in thlist:
            tag = self._db.get_tag_from_handle(tag_handle)
            tags.append(tag.get_name())
        for text in sorted(tags):
            self.doc.start_row()
            self.write_cell(text, span=2)
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_images(self):
        """write any images the person has"""

        media_list = self.person.get_media_list()
        if (not self.use_images) or (not media_list):
            return

        i_total = 0
        for media_ref in media_list:
            media_handle = media_ref.get_reference_handle()
            if media_handle:
                media = self._db.get_media_from_handle(media_handle)
                if media and media.get_mime_type():
                    if media.get_mime_type().startswith("image"):
                        i_total += 1
        if i_total == 0:
            return
        # if there is only one image, and it is the first Gallery item, it
        # will be shown up at the top, so there's no reason to show it here;
        # but if there's only one image and it is not the first Gallery
        # item (maybe the first is a PDF, say), then we need to show it
        if (i_total == 1) and self.mime0 and self.mime0.startswith("image"):
            return

        self.doc.start_table("images", "IDS-GalleryTable")
        cells = 3  # the GalleryTable has 3 cells
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", cells)
        self.write_paragraph(self._("Images"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()
        media_count = 0
        image_count = 0
        while media_count < len(media_list):
            media_ref = media_list[media_count]
            media_handle = media_ref.get_reference_handle()
            media = self._db.get_media_from_handle(media_handle)
            if media is None:
                self._user.notify_db_repair(
                    _("Non existing media found in the Gallery")
                )
                self.doc.end_table()
                self.doc.start_paragraph("IDS-Normal")
                self.doc.end_paragraph()
                return
            mime_type = media.get_mime_type()
            if not mime_type or not mime_type.startswith("image"):
                media_count += 1
                continue
            description = media.get_description()
            if image_count % cells == 0:
                self.doc.start_row()
            self.doc.start_cell("IDS-NormalCell")
            self.write_paragraph(description, style="IDS-ImageCaption")
            utils.insert_image(
                self._db,
                self.doc,
                media_ref,
                self._user,
                align="center",
                w_cm=5.0,
                h_cm=5.0,
            )
            self.do_attributes(
                media.get_attribute_list() + media_ref.get_attribute_list()
            )
            self.doc.end_cell()
            if image_count % cells == cells - 1:
                self.doc.end_row()
            media_count += 1
            image_count += 1
        if image_count % cells != 0:
            self.doc.end_row()
        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_families(self):
        """write any families the person has"""

        family_handle_list = self.person.get_family_handle_list()
        if not len(family_handle_list):
            return

        self.doc.start_table("three", "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._("Families"), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()
        self.doc.end_table()

        for family_handle in family_handle_list:
            self.doc.start_table("three", "IDS-IndTable")
            family = self._db.get_family_from_handle(family_handle)
            self.family_notes_list += family.get_note_list()
            if self.person.get_handle() == family.get_father_handle():
                spouse_id = family.get_mother_handle()
            else:
                spouse_id = family.get_father_handle()
            self.doc.start_row()
            self.doc.start_cell("IDS-NormalCell", 2)
            if spouse_id:
                spouse = self._db.get_person_from_handle(spouse_id)
                text = self.get_name(spouse)
                mark = utils.get_person_mark(self._db, spouse)
            else:
                spouse = None
                text = self._("unknown")
                mark = None
            endnotes = self._cite_endnote(family)
            self.write_paragraph(text, endnotes=endnotes, mark=mark, style="IDS-Spouse")
            self.doc.end_cell()
            self.doc.end_row()

            event_ref_list = family.get_event_ref_list()
            for event_ref, event in self.get_event_list(event_ref_list):
                self.write_fact(event_ref, event)

            child_ref_list = family.get_child_ref_list()
            if len(child_ref_list):
                self.doc.start_row()
                self.write_cell(self._("Children"))
                self.doc.start_cell("IDS-ListCell")
                for child_ref in child_ref_list:
                    child = self._db.get_person_from_handle(child_ref.ref)
                    name = self.get_name(child)
                    mark = utils.get_person_mark(self._db, child)
                    endnotes = self._cite_endnote(child_ref)
                    self.write_paragraph(name, endnotes=endnotes, mark=mark)
                self.doc.end_cell()
                self.doc.end_row()

            attr_list = family.get_attribute_list()
            if len(attr_list) and self.use_attrs:
                self.doc.start_row()
                self.write_cell(self._("Attributes"))
                self.doc.start_cell("IDS-ListCell")
                self.do_attributes(attr_list)
                self.doc.end_cell()
                self.doc.end_row()

            self.doc.end_table()

            ord_list = family.get_lds_ord_list()
            if len(ord_list):
                self.doc.start_table("ordinances2", "IDS-OrdinanceTable2")
                self.doc.start_row()
                self.write_cell(self._("LDS Ordinance"))
                self.write_cell(self._("Type"), style="IDS-TableSection")
                self.write_cell(self._("Date"), style="IDS-TableSection")
                self.write_cell(self._("Status"), style="IDS-TableSection")
                self.write_cell(self._("Temple"), style="IDS-TableSection")
                self.write_cell(self._("Place"), style="IDS-TableSection")
                self.doc.end_row()

                for lds_ord in ord_list:
                    otype = self._(lds_ord.type2str())
                    date = self._get_date(lds_ord.get_date_object())
                    status = self._(lds_ord.status2str())
                    temple = TEMPLES.name(lds_ord.get_temple())
                    place_name = ""
                    place_endnote = ""
                    place_handle = lds_ord.get_place_handle()
                    if place_handle:
                        place = self._db.get_place_from_handle(place_handle)
                        place_name = _pd.display_event(
                            self._db, lds_ord, self.place_format
                        )
                        place_endnote = self._cite_endnote(place)
                    endnotes = self._cite_endnote(lds_ord, prior=place_endnote)
                    self.doc.start_row()
                    self.write_cell("")
                    self.write_cell(otype, endnotes)
                    self.write_cell(date)
                    self.write_cell(status)
                    self.write_cell(temple)
                    self.write_cell(place_name)
                    self.doc.end_row()
                self.doc.end_table()

        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def get_event_list(self, event_ref_list):
        """
        Return a list of (EventRef, Event) pairs.  Order by event date
        if the user option is set.
        """
        event_list = []
        for ind, event_ref in enumerate(event_ref_list):
            if event_ref:
                event = self._db.get_event_from_handle(event_ref.ref)
                if event:
                    if event.get_type() == EventType.CENSUS and not self.use_census:
                        continue
                    sort_value = event.get_date_object().get_sort_value()
                    # first sort on date, equal dates, then sort as in GUI.
                    event_list.append(
                        (str(sort_value) + "%04i" % ind, event_ref, event)
                    )

        if self.sort:
            event_list.sort()

        return [(item[1], item[2]) for item in event_list]

    def write_section(self, event_ref_list, event_group_sect):
        """
        Writes events in a single event group.
        """
        self.doc.start_table(event_group_sect, "IDS-IndTable")
        self.doc.start_row()
        self.doc.start_cell("IDS-TableHead", 2)
        self.write_paragraph(self._(event_group_sect), style="IDS-SectionTitle")
        self.doc.end_cell()
        self.doc.end_row()

        for event_ref, event in self.get_event_list(event_ref_list):
            self.write_fact(event_ref, event)

        self.doc.end_table()
        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

    def write_events(self):
        """
        Write events.  The user can create separate sections for a
        pre-defined set of event groups.  When an event has a type
        contained within a group it is moved from the Individual Facts
        section into its own section.
        """
        event_dict = defaultdict(list)
        event_ref_list = self.person.get_event_ref_list()
        for event_ref in event_ref_list:
            if event_ref:
                event = self._db.get_event_from_handle(event_ref.ref)
                group = TYPE2GROUP[event.get_type().value]
                if _(group) not in self.section_list:
                    group = SECTION_LIST[0]
                event_dict[group].append(event_ref)

        # Write separate event group sections
        for group in SECTION_LIST:
            if group in event_dict:
                self.write_section(event_dict[group], group)

    def write_cell(self, text, endnotes=None, mark=None, style="IDS-Normal", span=1):
        """write a cell"""
        self.doc.start_cell("IDS-NormalCell", span)
        self.write_paragraph(text, endnotes=endnotes, mark=mark, style=style)
        self.doc.end_cell()

    def write_paragraph(self, text, endnotes=None, mark=None, style="IDS-Normal"):
        """write a paragraph"""
        self.doc.start_paragraph(style)
        self.doc.write_text(text, mark)
        if endnotes:
            self.doc.start_superscript()
            self.doc.write_text(endnotes)
            self.doc.end_superscript()
        self.doc.end_paragraph()

    def write_report(self):
        """write the report"""
        plist = self._db.get_person_handles(sort_handles=True, locale=self._locale)
        if self.filter:
            ind_list = self.filter.apply(self._db, plist, user=self._user)
        else:
            ind_list = plist
        if not ind_list:
            raise ReportError(_("Empty report"), _("You did not specify anybody"))

        if self._user:
            self._user.begin_progress(
                _("Complete Individual Report"), _("Generating report"), len(ind_list)
            )
        for count, person_handle in enumerate(ind_list):
            if self._user:
                self._user.step_progress()
            self.person = self._db.get_person_from_handle(person_handle)
            if self.person is None:
                continue
            self.family_notes_list = []
            self.names_notes_list = []
            self.write_person(count)
        if self._user:
            self._user.end_progress()

    def write_person(self, count):
        """write a person"""
        if count != 0:
            self.doc.page_break()
        self.bibli = Bibliography(Bibliography.MODE_DATE | Bibliography.MODE_PAGE)

        if self.name_is_title:
            title1 = self._name_display.display(self.person)
        else:
            title1 = self._("Complete Individual Report")
            text2 = self._name_display.display(self.person)
            mark2 = IndexMark(text2, INDEX_TYPE_TOC, 2)
        mark1 = IndexMark(title1, INDEX_TYPE_TOC, 1)
        self.doc.start_paragraph("IDS-Title")
        self.doc.write_text(title1, mark1)
        self.doc.end_paragraph()
        if not self.name_is_title:
            self.doc.start_paragraph("IDS-Title")
            self.doc.write_text(text2, mark2)
            self.doc.end_paragraph()

        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

        name = self.person.get_primary_name()
        text = self.get_name(self.person)
        mark = utils.get_person_mark(self._db, self.person)
        endnotes = self._cite_endnote(self.person)
        endnotes = self._cite_endnote(name, prior=endnotes)

        family_handle = self.person.get_main_parents_family_handle()
        if family_handle:
            family = self._db.get_family_from_handle(family_handle)
            father_inst_id = family.get_father_handle()
            if father_inst_id:
                father_inst = self._db.get_person_from_handle(father_inst_id)
                father = self.get_name(father_inst)
                fmark = utils.get_person_mark(self._db, father_inst)
            else:
                father = ""
                fmark = None
            mother_inst_id = family.get_mother_handle()
            if mother_inst_id:
                mother_inst = self._db.get_person_from_handle(mother_inst_id)
                mother = self.get_name(mother_inst)
                mmark = utils.get_person_mark(self._db, mother_inst)
            else:
                mother = ""
                mmark = None
        else:
            father = ""
            fmark = None
            mother = ""
            mmark = None

        media_list = self.person.get_media_list()
        p_style = "IDS-PersonTable2"
        self.mime0 = None
        if self.use_images and len(media_list) > 0:
            media0 = media_list[0]
            media_handle = media0.get_reference_handle()
            media = self._db.get_media_from_handle(media_handle)
            self.mime0 = media.get_mime_type()
            if self.mime0 and self.mime0.startswith("image"):
                image_filename = media_path_full(self._db, media.get_path())
                if os.path.exists(image_filename):
                    p_style = "IDS-PersonTable"  # this is tested for, also
                else:
                    self._user.warn(
                        _("Could not add photo to page"),
                        # Translators: for French, else ignore
                        _("%(str1)s: %(str2)s")
                        % {"str1": image_filename, "str2": _("File does not exist")},
                    )

        self.doc.start_table("person", p_style)
        self.doc.start_row()

        # Translators: needed for French, ignore otherwise
        ignore = self._("%s:")
        self.doc.start_cell("IDS-NormalCell")
        self.write_paragraph(self._("%s:") % self._("Name"))
        self.write_paragraph(self._("%s:") % self._("Gender"))
        self.write_paragraph(self._("%s:") % self._("Father"))
        self.write_paragraph(self._("%s:") % self._("Mother"))
        self.doc.end_cell()

        self.doc.start_cell("IDS-NormalCell")
        self.write_paragraph(text, endnotes, mark)
        if self.person.get_gender() == Person.MALE:
            self.write_paragraph(self._("Male"))
        elif self.person.get_gender() == Person.FEMALE:
            self.write_paragraph(self._("Female"))
        elif self.person.get_gender() == Person.OTHER:
            self.write_paragraph(self._("Other"))
        else:
            self.write_paragraph(self._("Unknown"))
        self.write_paragraph(father, mark=fmark)
        self.write_paragraph(mother, mark=mmark)
        self.doc.end_cell()

        if p_style == "IDS-PersonTable":
            self.doc.start_cell("IDS-NormalCell")
            self.doc.add_media(
                image_filename, "right", 4.0, 4.0, crop=media0.get_rectangle()
            )
            endnotes = self._cite_endnote(media0)
            attr_list = media0.get_attribute_list()
            if len(attr_list) == 0 or not self.use_attrs:
                text = _("(image)")
            else:
                for attr in attr_list:
                    attr_type = attr.get_type().type2base()
                    # Translators: needed for French, ignore otherwise
                    text = self._("%(str1)s: %(str2)s") % {
                        "str1": self._(attr_type),
                        "str2": attr.get_value(),
                    }
                    endnotes = self._cite_endnote(attr, prior=endnotes)
                    self.write_paragraph(
                        "(%s)" % text, endnotes=endnotes, style="IDS-ImageNote"
                    )
                    endnotes = ""
            if endnotes and len(attr_list) == 0:
                self.write_paragraph(text, endnotes=endnotes, style="IDS-ImageNote")
            self.doc.end_cell()

        self.doc.end_row()
        self.doc.end_table()

        self.doc.start_paragraph("IDS-Normal")
        self.doc.end_paragraph()

        self.write_home_relationship()
        self.write_alt_names()
        self.write_events()
        self.write_alt_parents()
        self.write_families()
        self.write_addresses()
        self.write_associations()
        self.write_attributes()
        self.write_LDS_ordinances()
        self.write_tags()
        self.write_images()
        self.write_note()
        if self.use_srcs:
            if self.use_pagebreak and self.bibli.get_citation_count():
                self.doc.page_break()
            Endnotes.write_endnotes(
                self.bibli,
                self._db,
                self.doc,
                printnotes=self.use_src_notes,
                elocale=self._locale,
            )

    def combine(self, format_both, format_single, str1, str2):
        """Combine two strings with a given format."""
        text = ""
        if str1 and str2:
            text = self._(format_both) % {"str1": str1, "str2": str2}
        elif str1 and not str2:
            text = format_single % str1
        elif str2 and not str1:
            text = format_single % str2
        return text

    def _cite_endnote(self, obj, prior=""):
        """cite any endnotes the person has"""
        if not self.use_srcs:
            return ""
        if not obj:
            return prior

        txt = Endnotes.cite_source(self.bibli, self._db, obj, self._locale)
        if not txt:
            return prior
        if prior:
            # Translators: needed for Arabic, ignore otherwise
            txt = self._("%(str1)s, %(str2)s") % {"str1": prior, "str2": txt}
        return txt

    def do_attributes(self, attr_list):
        """a convenience method"""
        if not self.use_attrs:
            return
        for attr in attr_list:
            attr_type = attr.get_type().type2base()
            # Translators: needed for French, ignore otherwise
            text = self._("%(str1)s: %(str2)s") % {
                "str1": self._(attr_type),
                "str2": attr.get_value(),
            }
            endnotes = self._cite_endnote(attr)
            self.write_paragraph(text, endnotes)


# ------------------------------------------------------------------------
#
# IndivCompleteOptions
#
# ------------------------------------------------------------------------
class IndivCompleteOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        self.__filter = None
        self.__cites = None
        self.__incsrcnotes = None
        self._nf = None
        self.__show_relships = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return self.__filter.get_filter().get_name()

    def add_menu_options(self, menu):
        ################################
        category_name = _("Report Options")
        ################################

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Select the filter to be applied to the report."))
        menu.add_option(category_name, "filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Filter Person"))
        self.__pid.set_help(_("The center person for the filter."))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        sort = BooleanOption(_("List events chronologically"), True)
        sort.set_help(_("Whether to sort events into chronological order."))
        menu.add_option(category_name, "sort", sort)

        pageben = BooleanOption(_("Page break before end notes"), False)
        pageben.set_help(_("Whether to start a new page before the end notes."))
        menu.add_option(category_name, "pageben", pageben)

        name_is_title = BooleanOption(_("Use name of person as title"), False)
        name_is_title.set_help(
            _(
                "Whether the title should be the name of the "
                "person, or 'Complete Individual Report'"
            )
        )
        menu.add_option(category_name, "name_title", name_is_title)

        ################################
        category_name = _("Report Options (2)")
        ################################

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        self.__update_filters()

        stdoptions.add_place_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

        ################################
        category_name = _("Include")
        ################################

        incl_notes = BooleanOption(_("Include Notes"), True)
        incl_notes.set_help(_("Whether to include Person and Family Notes."))
        menu.add_option(category_name, "incl_notes", incl_notes)

        self.__cites = BooleanOption(_("Include Source Information"), True)
        self.__cites.set_help(_("Whether to cite sources."))
        menu.add_option(category_name, "cites", self.__cites)
        self.__cites.connect("value-changed", self.__sources_changed)

        self.__incsrcnotes = BooleanOption(_("Include sources notes"), False)
        self.__incsrcnotes.set_help(
            _(
                "Whether to include source notes in the Endnotes section. "
                "Only works if Include sources is selected."
            )
        )
        menu.add_option(category_name, "incsrcnotes", self.__incsrcnotes)
        self.__incsrcnotes.connect("value-changed", self.__sources_changed)
        self.__sources_changed()

        images = BooleanOption(_("Include Photo/Images from Gallery"), True)
        images.set_help(_("Whether to include images."))
        menu.add_option(category_name, "images", images)

        ################################
        category_name = _("Include (2)")
        ################################

        stdoptions.add_gramps_id_option(menu, category_name)

        tags = BooleanOption(_("Include Tags"), True)
        tags.set_help(_("Whether to include tags."))
        menu.add_option(category_name, "incl_tags", tags)

        attributes = BooleanOption(_("Include Attributes"), True)
        attributes.set_help(_("Whether to include attributes."))
        menu.add_option(category_name, "incl_attrs", attributes)

        census = BooleanOption(_("Include Census Events"), True)
        census.set_help(_("Whether to include Census Events."))
        menu.add_option(category_name, "incl_census", census)

        self.__show_relships = BooleanOption(
            _("Include relationship to center person"), False
        )
        self.__show_relships.set_help(
            _("Whether to include relationships to the center person")
        )
        menu.add_option(category_name, "incl_relname", self.__show_relships)

        ################################
        category_name = _("Sections")
        ################################

        opt = BooleanListOption(_("Event groups"))
        opt.set_help(_("Check if a separate section is required."))
        for section in SECTION_LIST:
            if section != SECTION_LIST[0]:
                opt.add_button(_(section), True)

        menu.add_option(category_name, "sections", opt)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        nfv = self._nf.get_value()
        filter_list = utils.get_person_filters(
            person, include_single=True, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 1:  # "Entire Database" (as "include_single=True")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def __sources_changed(self):
        """
        If Endnotes are not enabled, disable sources in the Endnotes.
        """
        cites_value = self.__cites.get_value()
        self.__incsrcnotes.set_available(cites_value)

    def make_default_style(self, default_style):
        """Make the default output style for the Individual Complete Report."""
        # Paragraph Styles
        font = FontStyle()
        font.set_bold(1)
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(16)
        para = ParagraphStyle()
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_top_margin(utils.pt2cm(8))
        para.set_bottom_margin(utils.pt2cm(8))
        para.set_font(font)
        para.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("IDS-Title", para)

        font = FontStyle()
        font.set_bold(1)
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(12)
        font.set_italic(1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(3))
        para.set_bottom_margin(utils.pt2cm(3))
        para.set_description(_("The style used for the section headers."))
        default_style.add_paragraph_style("IDS-SectionTitle", para)

        font = FontStyle()
        font.set_bold(1)
        font.set_type_face(FONT_SANS_SERIF)
        font.set_size(12)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(3))
        para.set_bottom_margin(utils.pt2cm(3))
        para.set_description(_("The style used for the spouse's name."))
        default_style.add_paragraph_style("IDS-Spouse", para)

        font = FontStyle()
        font.set_size(12)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(3))
        para.set_bottom_margin(utils.pt2cm(3))
        para.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("IDS-Normal", para)

        font = FontStyle()
        font.set_size(12)
        font.set_italic(1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(3))
        para.set_bottom_margin(utils.pt2cm(3))
        para.set_description(_("The basic style used for table headings."))
        default_style.add_paragraph_style("IDS-TableSection", para)

        font = FontStyle()
        font.set_size(8)
        para = ParagraphStyle()
        para.set_alignment(PARA_ALIGN_RIGHT)
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(3))
        para.set_bottom_margin(utils.pt2cm(3))
        para.set_description(_("The style used for image notes."))
        default_style.add_paragraph_style("IDS-ImageNote", para)

        font = FontStyle()
        font.set_size(8)
        para = ParagraphStyle()
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(3))
        para.set_bottom_margin(utils.pt2cm(3))
        para.set_description(_("The style used for image descriptions."))
        default_style.add_paragraph_style("IDS-ImageCaption", para)

        # Table Styles
        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(2)
        tbl.set_column_width(0, 20)
        tbl.set_column_width(1, 80)
        default_style.add_table_style("IDS-IndTable", tbl)

        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(2)
        tbl.set_column_width(0, 50)
        tbl.set_column_width(1, 50)
        default_style.add_table_style("IDS-ParentsTable", tbl)

        cell = TableCellStyle()
        cell.set_top_border(1)
        cell.set_bottom_border(1)
        default_style.add_cell_style("IDS-TableHead", cell)

        cell = TableCellStyle()
        default_style.add_cell_style("IDS-NormalCell", cell)

        cell = TableCellStyle()
        cell.set_longlist(1)
        default_style.add_cell_style("IDS-ListCell", cell)

        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(3)
        tbl.set_column_width(0, 20)
        tbl.set_column_width(1, 50)
        tbl.set_column_width(2, 30)
        default_style.add_table_style("IDS-PersonTable", tbl)

        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(2)
        tbl.set_column_width(0, 20)
        tbl.set_column_width(1, 80)
        default_style.add_table_style("IDS-PersonTable2", tbl)

        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(5)
        tbl.set_column_width(0, 22)  # Type
        tbl.set_column_width(1, 22)  # Date
        tbl.set_column_width(2, 16)  # Status
        tbl.set_column_width(3, 22)  # Temple
        tbl.set_column_width(4, 18)  # Place
        default_style.add_table_style("IDS-OrdinanceTable", tbl)

        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(6)
        tbl.set_column_width(0, 20)  # empty
        tbl.set_column_width(1, 18)  # Type
        tbl.set_column_width(2, 18)  # Date
        tbl.set_column_width(3, 14)  # Status
        tbl.set_column_width(4, 18)  # Temple
        tbl.set_column_width(5, 12)  # Place
        default_style.add_table_style("IDS-OrdinanceTable2", tbl)

        tbl = TableStyle()
        tbl.set_width(100)
        tbl.set_columns(3)
        tbl.set_column_width(0, 33)
        tbl.set_column_width(1, 33)
        tbl.set_column_width(2, 34)
        default_style.add_table_style("IDS-GalleryTable", tbl)

        Endnotes.add_endnote_styles(default_style)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015       Craig Anderson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

""" Recursive base classes for reports
"""

from gramps.gen.plug.report import utils


# ------------------------------------------------------------------------
#
# Livrecurse base objects only
#
# ------------------------------------------------------------------------
class _PersonSeen:
    """librecurse base boject only
    Keep track of people that have been seen so we can call the correct
    virtual method.
    """

    def __init__(self):
        self.people_seen = set()

    def add_person(self, level, person_handle, family_handle):
        """a person is seen for the first time"""
        pass

    def add_person_again(self, level, person_handle, family_handle):
        """a person is seen again"""
        pass

    def _add_person(self, level, person_handle, family_handle):
        """Which virtual method to call?"""
        if person_handle is not None and person_handle in self.people_seen:
            self.add_person_again(level, person_handle, family_handle)
        else:
            self.add_person(level, person_handle, family_handle)
        if person_handle is not None:
            self.people_seen.add(person_handle)


class _FamilySeen:
    """librecurse base boject only
    Keep track of the famalies that have been seen so we can call the correct
    virtual method.
    """

    def __init__(self):
        self.families_seen = set()

    def add_marriage(self, level, person_handle, family_handle):
        """Makes a marriage"""
        pass

    def add_marriage_again(self, level, person_handle, family_handle):
        """Makes a marriage"""
        pass

    def _add_marriage(self, level, person_handle, family_handle):
        """Makes a marriage"""
        if family_handle in self.families_seen:
            self.add_marriage_again(level, person_handle, family_handle)
        else:
            self.add_marriage(level, person_handle, family_handle)
        self.families_seen.add(family_handle)


class _StopRecurse:
    """A simple class to break out the
    . stop_recursion
    . can_recurse
    . continue_recursion
    methods
    """

    def __init__(self):
        # The default value.  Lets recurse.
        self.__stop_recursion = False

    def stop_recursion(self):
        """Stop Recursion at theis person/family"""
        self.__stop_recursion = True

    def continue_recursion(self):
        """Used to allow recursion again"""
        self.__stop_recursion = False

    def can_recurse(self):
        """Has the upper class told up to stop or can we continue?"""
        return not self.__stop_recursion


# ------------------------------------------------------------------------
#
# Class DescendPerson
#
# ------------------------------------------------------------------------
class DescendPerson(_StopRecurse, _PersonSeen, _FamilySeen):
    """Recursive (down) base class

    The following methods need to be sub-classed as needed:
    . add_person
    . add_person_again (called when a person is seen a second or more times)
    if you don't want to see marriages don't subclass the following two
    . add_marriage
    . add_marriage_again (when a marriage is seen a second or more times)

    returns:
    . add_person, add_person_again, add_marriage, add_marriage_again return
    . . index -> a tuple in the form
    . . . generation, which generational level >= 1
    . . . level
    . . . . 0 = A direct child
    . . . . 1 = spouse of above (0)
    . . . . 2 = spouse of 1
    . . . . 3 = spouse of 2
    . . . . 4 etc
    . . person_handle
    . . family_handle

    Public variables:
    . families_seen a set of all famalies seen.
    . people_seen, a set of all people seen.
    . . useful for knowing if a recursion (kid marring a grandparent)
    . . has happened.
    These can be edited if needed
    . appending can be useful for excluding parts of the tree

    Methods (tools):
    is_direct_descendant - is this person a direct descendant
    . in the example 'kid 1 of mom and other spouse' is NOT
    stop_recursion - tells the recursion to stop going down
    . mostly used in add_person_again and add_marriage_again
    has_children - checks to see if the person:
    . is NOT already seen and has hildren.

    Methods (informative)
    . These are the methods that need to be subclassed
    . all methods are given the:
    . . level in (Generagion, Spousal level) tuple
    . . person_handle of the person
    . . family_handle of the family
    add_marriage
    add_marriage_again

    Virtual methods in PersonSeen
    . add_person - The recursion found a new person in the tree
    . add_person_again - found a person again
    . . a prolific person or recursion

    Methods (recursive)
    recurse - The main recursive routine. needs:
    . person_handle
    . g_level - Generation level of this person
    . . if max_gen is 2 and g_level is 1, only this generation
    . . will be displayed.
    . s_level - spousal level - most always 0
    recurse_parents - Thes same as above except:
    . mom (the spouse) is still shown even if s_level == 0
    . . father will have a level of (g_level,0), mother (g_level, 1)
    """

    def __init__(self, dbase, maxgen, maxspouse=0):
        _PersonSeen.__init__(self)
        _FamilySeen.__init__(self)
        _StopRecurse.__init__(self)

        """ initalized with the
            . database
            . maxgen is the max generations (down) of people to return
            . maxspouse is the level of spouses to recruse through
            . . 0 = no spouses, 1 = spouses of a direct descendant
            . . 2 = spouses of 1, 3 = spouses of 2, etc.  See example below

        """
        # example:  maxgen = 2, maxspouses = 2
        # (1,0) father
        #   (1,1) Mother
        #     (1,2) Mothers other spouse
        #       (2,0) kid 1 of mom and other spouse
        #     (2,0) Kid 1 of father and mother
        #   (1,1) fathers other spouse
        #     (2,0) Kid 1 of father and fathers other spouse
        #       (2,1) Spouse of Kid 1 of father and fathers other spouse

        self.database = dbase

        assert maxgen > 0
        self.max_generations = maxgen

        assert maxspouse >= 0
        self.max_spouses = maxspouse

        # can we bold direct descendants?
        # bold_now will have only three values
        # 0 - no bolding
        # 1 - Only bold the first person
        # 2 - Bold all direct descendants
        self.__bold_now = 1
        self.__this_slevel = -1

    def is_direct_descendant(self):
        """Is this person a direct descendant?
        . Can we bold this perosn and
        . are they a direct child of the father/mother
        . . not a spouse
        """
        return self.__bold_now != 0 and self.__this_slevel == 0

    def has_children(self, person_handle):
        """
        Quickly check to see if this person has children
        still we want to respect the people_seen list
        """

        if not person_handle or person_handle in self.people_seen:
            return False

        person = self.database.get_person_from_handle(person_handle)

        for family_handle in person.get_family_handle_list():
            if family_handle not in self.families_seen:
                family = self.database.get_family_from_handle(family_handle)

                if family.get_child_ref_list():
                    return True
        return False

    def recurse(self, person_handle, g_level, s_level):
        """traverse the descendants recursively
        until either the end of a line is found,
        or until we reach the maximum number of generations
        or we reach the max number of spouses
        that we want to deal with"""

        if not person_handle:
            return
        if g_level > self.max_generations:
            return  # one generation too many
        if s_level > 0 and s_level == self.max_spouses:
            return
        # if person_handle in self.people_seen: return

        person = self.database.get_person_from_handle(person_handle)
        family_handles = person.get_family_handle_list()
        if s_level == 0:
            val = family_handles[0] if family_handles else None
            self.__this_slevel = s_level
            self._add_person((g_level, s_level), person_handle, val)

            if self.__bold_now == 1:
                self.__bold_now = 0

            if not self.can_recurse():
                self.continue_recursion()
                return

        if s_level == 1:
            tmp_bold = self.__bold_now
            self.__bold_now = 0

        for family_handle in family_handles:
            # Marriage box if the option is there.
            self._add_marriage((g_level, s_level + 1), person_handle, family_handle)

            if not self.can_recurse():
                self.continue_recursion()
                return

            family = self.database.get_family_from_handle(family_handle)

            spouse_handle = utils.find_spouse(person, family)
            if self.max_spouses > s_level:
                self.__this_slevel = s_level + 1
                self._add_person((g_level, s_level + 1), spouse_handle, family_handle)

                if not self.can_recurse:
                    self.continue_recursion()
                    return

            mykids = [kid.ref for kid in family.get_child_ref_list()]

            if self.can_recurse():
                for child_ref in mykids:
                    self.recurse(child_ref, g_level + 1, 0)
            else:
                self.continue_recursion()

            if self.max_spouses > s_level:
                # spouse_handle = utils.find_spouse(person,family)
                self.recurse(spouse_handle, g_level, s_level + 1)

        if s_level == 1:
            self.__bold_now = tmp_bold

    def recurse_parents(self, family_handle, g_level):
        """
        Adds a family.
        ignoring maxspouse, s_level assumed 0 and 1
        father is (g_level,0) and mother is (g_level,1)
        children are (g_level+1,0) and respects maxgen
        """

        if family_handle is None:
            return

        family = self.database.get_family_from_handle(family_handle)
        father_h = family.get_father_handle()
        mother_h = family.get_mother_handle()

        self.__bold_now = 2
        self.__this_slevel = 0
        # if father_h:
        father_b = self._add_person((g_level, 0), father_h, family_handle)
        # else:
        #    #TODO - should send family_h instead of None?
        #    father_b = self._add_person((g_level, 0), None, family_h)
        # self.people_seen.add(father_h)

        family_b = self._add_marriage((g_level, 1), father_h, family_handle)

        self.__bold_now = 0
        self.__this_slevel = 1
        mother_b = self._add_person((g_level, 1), mother_h, family_handle)

        self.__bold_now = 2
        for child_ref in family.get_child_ref_list():
            self.recurse(child_ref.ref, g_level + 1, 0)

        self.__bold_now = 0

        return (father_b, family_b, mother_b)

    def recurse_if(self, person_handle, g_level):
        """
        Quickly check to see if we want to continue recursion
        we still we want to respect the FamiliesSeen list
        """

        person = self.database.get_person_from_handle(person_handle)

        show = False
        myfams = person.get_family_handle_list()
        if len(myfams) > 1:  # and self.max_spouses > 0
            show = True

        # if self.max_spouses == 0 and not self.has_children(person_handle):
        #    self.people_seen.add(person_handle)
        #    show = False

        if show:
            self.__bold_now = 1
            self.recurse(person_handle, g_level, 0)


# ------------------------------------------------------------------------
#
# Class AscendPerson
#
# ------------------------------------------------------------------------
class AscendPerson(_StopRecurse, _PersonSeen):
    """Recursive (up) base class

    The following methods need to be sub-classed as needed:
    . add_person
    . add_person_again (called when a person is seen a second or more times)
    if you don't want to see marriages don't subclass the following
    . add_marriage
    . . index (below) will be the same as the father

    returns:
    . add_person, add_person_again, add_marriage all return
    . . index -> a tuple in the form
    . . . generation, which generational level >= 1
    . . . . Center person is 1
    . . . . Father/Mother is the generational level of the child + 1
    . . . index
    . . . . The center person is 1
    . . . . A father is the  index of the child * 2
    . . . . A mother is the (index of the child * 2) + 1
    . . person_handle  (May be None)
    . . family_handle  (May be None)

    Public variables:
    . people_seen, a set of all people seen.
    . . useful for knowing if a recursion (kid marring a grandparent)
    . . has happened.
    These can be edited if needed
    . people_seen, a set of all people seen.
    . . appending can be useful for excluding parts of the tree
    """

    def __init__(self, dbase, maxgen, maxfill=0):
        _PersonSeen.__init__(self)
        _StopRecurse.__init__(self)

        """ initalized with the
            . database
            . maxgen is the max generations (up) of people to return
            . . maxgen >= 1.  1 will only be the person.
            . maxfill is the max generations of blank (null) people to return
            . . maxfil >= 0.  0 (default) is no empty generations
        """
        self.database = dbase
        assert maxgen > 0
        self.max_generations = maxgen
        assert maxfill >= 0
        self.fill_out = maxfill

    def add_marriage(self, index, indi_handle, fams_handle):
        """Makes a marriage box and add that person into the Canvas."""
        # We are not using add_marriage only and not add_marriage_again
        # because the father will do any _again stuff if needed.
        pass

    def __fill(self, generation, index, mx_fill):
        """
        A skeleton of __iterate as person_handle == family_handle == None
        """
        if generation > self.max_generations or mx_fill == 0:
            # Gone too far.
            return

        self.add_person((generation, index), None, None)

        if not self.can_recurse():
            self.continue_recursion()
            return

        # Recursively call the function. It is okay if the handle is None,
        # since routine handles a handle of None

        self.__fill(generation + 1, index * 2, mx_fill - 1)
        if mx_fill > 1:  # marriage of parents
            self.add_marriage((generation + 1, index * 2), None, None)
            if not self.can_recurse():
                self.continue_recursion()
                return
        self.__fill(generation + 1, (index * 2) + 1, mx_fill - 1)

    def __iterate(self, generation, index, person_handle, full_family_handle):
        """
        Recursive function to walk back all parents of the current person.
        When max_generations are hit, we stop the traversal.

        Code pilfered from gramps/plugins/textreports/ancestorreport.py
        """

        # check for end of the current recursion level. This happens
        # if the person handle is None, or if the max_generations is hit

        if generation > self.max_generations:  # too many generations
            return

        if person_handle is None:  # Ran out of people
            return self.__fill(generation, index, self.fill_out)

        # retrieve the Person instance from the database from the
        # passed person_handle and find the parents from the list.
        # Since this report is for natural parents (birth parents),
        # we have to handle that parents may not
        person = self.database.get_person_from_handle(person_handle)

        # we have a valid person, add him/her
        self._add_person((generation, index), person_handle, full_family_handle)

        # has the user canceled recursion?
        if not self.can_recurse():
            self.continue_recursion()
            return

        # Now recurse on the parents
        family_handle = person.get_main_parents_family_handle()

        if family_handle is not None:
            family = self.database.get_family_from_handle(family_handle)
            father_handle = family.get_father_handle()
            mother_handle = family.get_mother_handle()
        else:
            father_handle = None
            mother_handle = None

        # Recursively call the function. It is okay if the handle is None,
        self.__iterate(
            generation + 1, index * 2, father_handle, family_handle
        )  # recurse on dad
        if generation < self.max_generations:
            if father_handle is not None:  # Stil winin max_generations
                self.add_marriage(
                    (generation + 1, index * 2), father_handle, family_handle
                )
            elif mother_handle is not None:
                self.add_marriage(
                    (generation + 1, index * 2), mother_handle, family_handle
                )
            elif family_handle is not None:
                self.add_marriage((generation + 1, index * 2), None, family_handle)
            elif self.fill_out > 0:
                self.add_marriage((generation + 1, index * 2), None, None)

            if not self.can_recurse():
                self.continue_recursion()
                return
        self.__iterate(
            generation + 1, (index * 2) + 1, mother_handle, family_handle
        )  # recurse mom

    def recurse(self, person_handle):
        """
        A simple header to make sure we pass in the correct information
        """
        person = self.database.get_person_from_handle(person_handle)
        return self.__iterate(
            1, 1, person_handle, person.get_main_parents_family_handle()
        )


# ------------
# Jer 29:11: "For I know the plans I have for you," declares the LORD,
# "plans to prosper you and not to harm you, plans to give you hope
# and a future."



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
# Copyright (C) 2014       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ..dialog import ErrorDialog
from ..managedwindow import ManagedWindow
from .displaytabs import GrampsTab
from gramps.gen.config import config
from ..dbguielement import DbGUIElement

# -------------------------------------------------------------------------
#
# Classes
#
# -------------------------------------------------------------------------


class RefTab(GrampsTab):
    """
    This class provides a simple tabpage for use on EditReference
    """

    def __init__(self, dbstate, uistate, track, name, widget):
        """
        @param dbstate: The database state. Contains a reference to
        the database, along with other state information. The GrampsTab
        uses this to access the database and to pass to and created
        child windows (such as edit dialogs).
        @type dbstate: DbState
        @param uistate: The UI state. Used primarily to pass to any created
        subwindows.
        @type uistate: DisplayState
        @param track: The window tracking mechanism used to manage windows.
        This is only used to pass to generted child windows.
        @type track: list
        @param name: Notebook label name
        @type name: str/unicode
        @param widget: widget to be shown in the tab
        @type widge: gtk widget
        """
        GrampsTab.__init__(self, dbstate, uistate, track, name)
        eventbox = Gtk.EventBox()
        eventbox.add(widget)
        self.pack_start(eventbox, True, True, 0)
        self._set_label(show_image=False)
        widget.connect("key_press_event", self.key_pressed)
        self.show_all()

    def is_empty(self):
        """
        Override base class
        """
        return False


# -------------------------------------------------------------------------
#
# EditReference class
#
# -------------------------------------------------------------------------
class EditReference(ManagedWindow, DbGUIElement):
    def __init__(self, state, uistate, track, source, source_ref, update):
        self.db = state.db
        self.dbstate = state
        self.uistate = uistate
        self.source_ref = source_ref
        self.source = source
        self.source_added = False
        self.update = update
        self.warn_box = None
        self.__tabs = []

        ManagedWindow.__init__(self, uistate, track, source_ref)
        DbGUIElement.__init__(self, self.db)

        self._local_init()
        self._set_size()
        self._create_tabbed_pages()
        self._setup_fields()
        self._connect_signals()
        self.show()
        self._post_init()

    def _local_init(self):
        """
        Derived class should do any pre-window initalization in this task.
        """
        pass

    def define_warn_box(self, box):
        self.warn_box = box

    def enable_warnbox(self):
        self.warn_box.show()

    def define_expander(self, expander):
        expander.set_expanded(True)
        expander.connect("activate", self.__on_expand)

    def __on_expand(self, expander):
        """
        Sets the packing of the expander widget to depend on whether or not
        it is expanded.
        """
        state = not expander.get_expanded()
        parent = expander.get_parent()
        parent.set_child_packing(expander, state, state, 0, Gtk.PackType.START)
        expander.set_vexpand(state)

    def _post_init(self):
        """
        Derived class should do any post-window initalization in this task.
        """
        pass

    def _setup_notebook_tabs(self, notebook):
        for child in notebook.get_children():
            label = notebook.get_tab_label(child)
            page_no = notebook.page_num(child)
            label.drag_dest_set(0, [], 0)
            label.connect("drag_motion", self._switch_page_on_dnd, notebook, page_no)
            child.set_parent_notebook(notebook)
        notebook.connect("key-press-event", self.key_pressed, notebook)

    def key_pressed(self, obj, event, notebook):
        """
        Handles the key being pressed on the notebook, pass to key press of
        current page.
        """
        pag = notebook.get_current_page()
        if not pag == -1:
            notebook.get_nth_page(pag).key_pressed(obj, event)

    def _switch_page_on_dnd(self, widget, context, x, y, time, notebook, page_no):
        if notebook.get_current_page() != page_no:
            notebook.set_current_page(page_no)

    def _add_tab(self, notebook, page):
        self.__tabs.append(page)
        notebook.insert_page(page, page.get_tab_widget(), -1)
        page.label.set_use_underline(True)
        return page

    def _connect_signals(self):
        pass

    def _setup_fields(self):
        pass

    def _create_tabbed_pages(self):
        pass

    def build_window_key(self, sourceref):
        # the window key for managedwindow identification. No need to return None
        if self.source and self.source.get_handle():
            return self.source.get_handle()
        else:
            return id(self)

    def define_ok_button(self, button, function):
        button.connect("clicked", function)
        button.set_sensitive(not self.db.readonly)

    def define_cancel_button(self, button):
        button.connect("clicked", self.close_and_cancel)

    def close_and_cancel(self, obj):
        self.close(obj)

    def check_for_close(self, handles):
        """
        Callback method for delete signals.
        If there is a delete signal of the primary object we are editing, the
        editor (and all child windows spawned) should be closed
        """
        if self.source.get_handle() in handles:
            self.close()

    def define_help_button(self, button, webpage="", section=""):
        from ..display import display_help

        button.connect("clicked", lambda x: display_help(webpage, section))
        button.set_sensitive(True)

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        for tab in self.__tabs:
            if hasattr(tab, "_cleanup_on_exit"):
                tab._cleanup_on_exit()
        self.__tabs = None
        self.dbstate = None
        self.uistate = None
        self.source_ref = None
        self.source = None
        self.update = None
        self.warn_box = None
        self.db = None
        self.callman.database = None
        self.callman = None

    def close(self, *obj):
        self._cleanup_db_connects()
        self._cleanup_connects()
        ManagedWindow.close(self)
        self._cleanup_on_exit()

    def _cleanup_db_connects(self):
        """
        All connects that happened to signals of the db must be removed on
        closed. This implies two things:
        1. The connects on the main view must be disconnected
        2. Connects done in subelements must be disconnected
        """
        # cleanup callbackmanager of this editor
        self._cleanup_callbacks()
        for tab in [tab for tab in self.__tabs if hasattr(tab, "callman")]:
            tab._cleanup_callbacks()

    def _cleanup_connects(self):
        """
        Connects to interface elements to things outside the element should be
        removed before destroying the interface
        """
        self._cleanup_local_connects()
        for tab in [
            tab for tab in self.__tabs if hasattr(tab, "_cleanup_local_connects")
        ]:
            tab._cleanup_local_connects()

    def _cleanup_local_connects(self):
        """
        Connects to interface elements to things outside the element should be
        removed before destroying the interface. This methods cleans connects
        of the main interface, not of the displaytabs.
        """
        pass

    def check_for_duplicate_id(self, type):
        """
        check to see if the gramps ID (if any) already exists

        type    : the gramps primary object type, a string
        returns : True if the gramps ID already exists, else False

        N.B. the various strings, string variables, and titles existed already
        """
        new_id = self.source.get_gramps_id()
        if new_id:
            id_func = getattr(self.db, "get_%s_from_gramps_id" % type.lower())
            old_primary = id_func(new_id)
            if old_primary:
                description = None
                if type == "Event":
                    msg1 = _("Cannot save event. ID already exists.")
                    description = old_primary.get_description()
                elif type == "Media":
                    msg1 = _("Cannot save media object. ID already exists.")
                    description = old_primary.get_description()
                elif type == "Repository":
                    msg1 = _("Cannot save repository. ID already exists.")
                    description = old_primary.get_name()
                else:
                    msg1 = _("Cannot save item. ID already exists.")
                if description:
                    msg2 = _(
                        "You have attempted to use the existing Gramps "
                        "ID with value %(id)s. This value is already "
                        "used by '%(prim_object)s'. Please enter a "
                        "different ID or leave blank to get the next "
                        "available ID value."
                    ) % {"id": new_id, "prim_object": description}
                else:
                    msg2 = _(
                        "You have attempted to use the existing Gramps "
                        "ID with value %(id)s. This value is already "
                        "used. Please enter a "
                        "different ID or leave blank to get the next "
                        "available ID value."
                    ) % {"id": new_id}
                ErrorDialog(msg1, msg2, parent=self.window)
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.const import URL_MANUAL_SECT2
from gramps.gen.lib import Event, NoteType
from gramps.gen.db import DbTxn
from ..display import display_help
from .editprimary import EditPrimary
from .objectentries import PlaceEntry
from ..glade import Glade
from ..dialog import ErrorDialog
from .displaytabs import (
    CitationEmbedList,
    NoteTab,
    GalleryTab,
    EventBackRefList,
    EventAttrEmbedList,
)
from ..widgets import (
    MonitoredEntry,
    PrivacyButton,
    MonitoredDataType,
    MonitoredDate,
    MonitoredTagList,
)
from gramps.gen.utils.db import get_participant_from_event

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("New_Event_dialog", "manual")


# -------------------------------------------------------------------------
#
# EditEvent class
#
# -------------------------------------------------------------------------
class EditEvent(EditPrimary):
    def __init__(self, dbstate, uistate, track, event, callback=None):
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            event,
            dbstate.db.get_event_from_handle,
            dbstate.db.get_event_from_gramps_id,
            callback,
        )

        self._init_event()

    def _init_event(self):
        if not self.db.readonly:
            self.commit_event = self.db.commit_event

    def empty_object(self):
        return Event()

    def get_menu_title(self):
        handle = self.obj.get_handle()
        if handle:
            who = get_participant_from_event(self.db, handle)
            desc = self.obj.get_description()
            event_name = self.obj.get_type()
            if desc:
                event_name = "%s - %s" % (event_name, desc)
            if who:
                event_name = "%s - %s" % (event_name, who)
            dialog_title = _("Event: %s") % event_name
        else:
            dialog_title = _("New Event")
        return dialog_title

    def get_custom_events(self):
        return sorted(self.dbstate.db.get_event_types(), key=lambda s: s.lower())

    def _local_init(self):
        self.top = Glade()
        self.set_window(self.top.toplevel, None, self.get_menu_title())
        self.setup_configs("interface.event", 600, 450)

        self.place = self.top.get_object("place")
        self.share_btn = self.top.get_object("select_place")
        self.add_del_btn = self.top.get_object("add_del_place")

    def _connect_signals(self):
        self.top.get_object("button111").connect("clicked", self.close)
        self.top.get_object("button126").connect("clicked", self.help_clicked)

        self.ok_button = self.top.get_object("ok")
        self.ok_button.set_sensitive(not self.db.readonly)
        self.ok_button.connect("clicked", self.save)

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("event-rebuild", self._do_close)
        self._add_db_signal("event-delete", self.check_for_close)
        self._add_db_signal("place-delete", self.place_delete)
        self._add_db_signal("place-update", self.place_update)

    def _setup_fields(self):
        # place, select_place, add_del_place

        self.place_field = PlaceEntry(
            self.dbstate,
            self.uistate,
            self.track,
            self.top.get_object("place"),
            self.top.get_object("place_event_box"),
            self.obj.set_place_handle,
            self.obj.get_place_handle,
            self.add_del_btn,
            self.share_btn,
        )

        self.descr_field = MonitoredEntry(
            self.top.get_object("event_description"),
            self.obj.set_description,
            self.obj.get_description,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

        self.event_menu = MonitoredDataType(
            self.top.get_object("personal_events"),
            self.obj.set_type,
            self.obj.get_type,
            custom_values=self.get_custom_events(),
        )

        self.date_field = MonitoredDate(
            self.top.get_object("date_entry"),
            self.top.get_object("date_stat"),
            self.obj.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """
        notebook = Gtk.Notebook()

        self.citation_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "event_editor_citations",
            self.get_menu_title(),
        )
        self._add_tab(notebook, self.citation_list)

        self.note_list = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "event_editor_notes",
            notetype=NoteType.EVENT,
        )
        self._add_tab(notebook, self.note_list)

        self.gallery_list = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_media_list(),
        )
        self._add_tab(notebook, self.gallery_list)

        self.attr_list = EventAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_attribute_list(),
            "event_editor_attributes",
        )
        self._add_tab(notebook, self.attr_list)

        handle_list = self.dbstate.db.find_backlink_handles(self.obj.handle)
        self.backref_list = EventBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            handle_list,
            "event_editor_references",
        )
        self._add_tab(notebook, self.backref_list)

        self._setup_notebook_tabs(notebook)

        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

        self.track_ref_for_deletion("citation_list")
        self.track_ref_for_deletion("note_list")
        self.track_ref_for_deletion("gallery_list")
        self.track_ref_for_deletion("attr_list")
        self.track_ref_for_deletion("backref_list")

    def build_menu_names(self, event):
        return (_("Edit Event"), self.get_menu_title())

    def help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def save(self, *obj):
        self.ok_button.set_sensitive(False)
        if self.object_is_empty():
            ErrorDialog(
                _("Cannot save event"),
                _(
                    "No data exists for this event. Please "
                    "enter data or cancel the edit."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        (uses_dupe_id, id) = self._uses_duplicate_id()
        if uses_dupe_id:
            prim_object = self.get_from_gramps_id(id)
            name = prim_object.get_description()
            msg1 = _("Cannot save event. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used by '"
                "%(prim_object)s'. Please enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": id, "prim_object": name}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        t = self.obj.get_type()
        if t.is_custom() and str(t) == "":
            ErrorDialog(
                _("Cannot save event"),
                _("The event type cannot be empty"),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        if not self.obj.handle:
            with DbTxn(
                _("Add Event (%s)") % self.obj.get_gramps_id(), self.db
            ) as trans:
                self.db.add_event(self.obj, trans)
        else:
            if self.data_has_changed():
                with DbTxn(
                    _("Edit Event (%s)") % self.obj.get_gramps_id(), self.db
                ) as trans:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_event_gramps_id())
                    self.db.commit_event(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj)

    def data_has_changed(self):
        """
        A date comparison can fail incorrectly because we have made the
        decision to store entered text in the date. However, there is no
        entered date when importing from a XML file, so we can get an
        incorrect fail.
        """

        if self.db.readonly:
            return False
        elif self.obj.handle:
            orig = self.get_from_handle(self.obj.handle)
            if orig:
                cmp_obj = orig
            else:
                cmp_obj = self.empty_object()
            return cmp_obj.serialize(True)[1:] != self.obj.serialize(True)[1:]
        else:
            cmp_obj = self.empty_object()
            return cmp_obj.serialize(True)[1:] != self.obj.serialize()[1:]

    def place_update(self, hndls):
        """Place changed outside of dialog, update text if its ours"""
        handle = self.obj.get_place_handle()
        if handle and handle in hndls:
            place = self.db.get_place_from_handle(handle)
            p_lbl = "%s [%s]" % (place.get_title(), place.gramps_id)
            self.top.get_object("place").set_text(p_lbl)

    def place_delete(self, hndls):
        """Place deleted outside of dialog, remove it if its ours"""
        handle = self.obj.get_place_handle()
        if handle and handle in hndls:
            self.obj.set_place_handle(None)
            self.top.get_object("place").set_markup(self.place_field.EMPTY_TEXT)
            self.place_field.set_button(False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Event object for Gramps.
"""

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .primaryobj import PrimaryObject
from .citationbase import CitationBase
from .notebase import NoteBase
from .mediabase import MediaBase
from .attrbase import AttributeBase
from .datebase import DateBase
from .placebase import PlaceBase
from .tagbase import TagBase
from .eventtype import EventType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

LOG = logging.getLogger(".citation")


# -------------------------------------------------------------------------
#
# Event class
#
# -------------------------------------------------------------------------
class Event(
    CitationBase, NoteBase, MediaBase, AttributeBase, DateBase, PlaceBase, PrimaryObject
):
    """
    The Event record is used to store information about some type of
    action that occurred at a particular place at a particular time,
    such as a birth, death, or marriage.

    A possible definition: Events are things that happen at some point in time
    (that we may not know precisely, though), at some place, may involve
    several people (witnesses, officers, notaries, priests, etc.) and may
    of course have sources, notes, media, etc.
    Compare this with attribute: :class:`~.attribute.Attribute`
    """

    def __init__(self, source=None):
        """
        Create a new Event instance, copying from the source if present.

        :param source: An event used to initialize the new event
        :type source: Event
        """

        PrimaryObject.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        MediaBase.__init__(self, source)
        AttributeBase.__init__(self)
        DateBase.__init__(self, source)
        PlaceBase.__init__(self, source)

        if source:
            self.__description = source.description
            self.__type = EventType(source.type)
        else:
            self.__description = ""
            self.__type = EventType()

    def serialize(self, no_text_date=False):
        """
        Convert the data held in the event to a Python tuple that
        represents all the data elements.

        This method is used to convert the object into a form that can easily
        be saved to a database.

        These elements may be primitive Python types (string, integers),
        complex Python types (lists or tuples, or Python objects. If the
        target database cannot handle complex types (such as objects or
        lists), the database is responsible for converting the data into
        a form that it can use.

        :returns: Returns a python tuple containing the data that should
                  be considered persistent.
        :rtype: tuple
        """
        return (
            self.handle,
            self.gramps_id,
            self.__type.serialize(),
            DateBase.serialize(self, no_text_date),
            self.__description,
            self.place,
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            MediaBase.serialize(self),
            AttributeBase.serialize(self),
            self.change,
            TagBase.serialize(self),
            self.private,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        from .attribute import Attribute
        from .date import Date
        from .mediaref import MediaRef

        return {
            "type": "object",
            "title": _("Event"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {"type": "string", "maxLength": 50, "title": _("Handle")},
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "type": EventType.get_schema(),
                "date": {
                    "oneOf": [{"type": "null"}, Date.get_schema()],
                    "title": _("Date"),
                },
                "description": {"type": "string", "title": _("Description")},
                "place": {
                    "type": ["string", "null"],
                    "maxLength": 50,
                    "title": _("Place"),
                },
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "media_list": {
                    "type": "array",
                    "items": MediaRef.get_schema(),
                    "title": _("Media"),
                },
                "attribute_list": {
                    "type": "array",
                    "items": Attribute.get_schema(),
                    "title": _("Attributes"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in an Event structure.

        :param data: tuple containing the persistent data associated with the
                     Event object
        :type data: tuple
        """
        (
            self.handle,
            self.gramps_id,
            the_type,
            date,
            self.__description,
            self.place,
            citation_list,
            note_list,
            media_list,
            attribute_list,
            self.change,
            tag_list,
            self.private,
        ) = data

        self.__type = EventType()
        self.__type.unserialize(the_type)
        DateBase.unserialize(self, date)
        MediaBase.unserialize(self, media_list)
        AttributeBase.unserialize(self, attribute_list)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        TagBase.unserialize(self, tag_list)
        return self

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle of
                  this object type.
        :rtype: bool
        """
        if classname == "Place":
            return self.place == handle
        return False

    def _remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str
        """
        if classname == "Place" and self.place in handle_list:
            self.place = ""

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        if classname == "Place" and self.place == old_handle:
            self.place = new_handle

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.__description, str(self.__type), self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.media_list + self.attribute_list

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return self.media_list + self.attribute_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return self.media_list + self.attribute_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        ret = (
            self.get_referenced_note_handles()
            + self.get_referenced_citation_handles()
            + self.get_referenced_tag_handles()
        )
        if self.place:
            ret.append(("Place", self.place))
        return ret

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.get_citation_child_list()

    def is_empty(self):
        """
        Return True if the Event is an empty object (no values set).

        :returns: True if the Event is empty
        :rtype: bool
        """
        date = self.get_date_object()
        place = self.get_place_handle()
        description = self.__description
        the_type = self.__type
        return (
            the_type == EventType.CUSTOM
            and date.is_empty()
            and not place
            and not description
        )

    def are_equal(self, other):
        """
        Return True if the passed Event is equivalent to the current Event.

        :param other: Event to compare against
        :type other: Event
        :returns: True if the Events are equal
        :rtype: bool
        """
        if other is None:
            other = Event(None)

        if (
            self.__type != other.type
            or ((self.place or other.place) and (self.place != other.place))
            or self.__description != other.description
            or self.private != other.private
            or (not self.get_date_object().is_equal(other.get_date_object()))
            or len(self.get_citation_list()) != len(other.get_citation_list())
        ):
            return False

        index = 0
        olist = other.get_citation_list()
        for handle in self.get_citation_list():
            # see comment in srefs_are_equal in gen/plug/report/_bibliography.py
            if handle != olist[index]:
                return False
            index += 1

        return True

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this event.

        Lost: handle, id, marker, type, date, place, description of acquisition.

        :param acquisition: The event to merge with the present event.
        :type acquisition: Event
        """
        self._merge_privacy(acquisition)
        self._merge_attribute_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_media_list(acquisition)
        self._merge_tag_list(acquisition)

    def set_type(self, the_type):
        """
        Set the type of the Event to the passed (int,str) tuple.

        :param the_type: Type to assign to the Event
        :type the_type: tuple
        """
        self.__type.set(the_type)

    def get_type(self):
        """
        Return the type of the Event.

        :returns: Type of the Event
        :rtype: tuple
        """
        return self.__type

    type = property(get_type, set_type, None, "Returns or sets type of the event")

    def set_description(self, description):
        """
        Set the description of the Event to the passed string.

        The string may contain any information.

        :param description: Description to assign to the Event
        :type description: str
        """
        self.__description = description

    def get_description(self):
        """
        Return the description of the Event.

        :returns: Returns the description of the Event
        :rtype: str
        """
        return self.__description

    description = property(
        get_description,
        set_description,
        None,
        "Returns or sets description of the event",
    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Gramps
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen.filters.rules/Person/_HasNameOriginType.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....lib.nameorigintype import NameOriginType


# -------------------------------------------------------------------------
#
# HasNameOriginType
#
# -------------------------------------------------------------------------
class HasNameOriginType(Rule):
    """Rule that checks the type of Surname origin"""

    labels = [_("Surname origin type:")]
    name = _("People with the <Surname origin type>")
    description = _("Matches people with a surname origin")
    category = _("General filters")

    def apply(self, db, person):
        if not self.list[0]:
            return False
        for name in [person.get_primary_name()] + person.get_alternate_names():
            for surname in name.get_surname_list():
                specified_type = NameOriginType()
                specified_type.set_from_xml_str(self.list[0])
                if surname.get_origintype() == specified_type:
                    return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2012       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Package providing filtering framework for Gramps.
"""

# ------------------------------------------------------------------------
#
# Gramps imports
#
# ------------------------------------------------------------------------
from ..lib.person import Person
from ..lib.family import Family
from ..lib.src import Source
from ..lib.citation import Citation
from ..lib.event import Event
from ..lib.place import Place
from ..lib.repo import Repository
from ..lib.media import Media
from ..lib.note import Note
from ..lib.tag import Tag
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# GenericFilter
#
# -------------------------------------------------------------------------
class GenericFilter:
    """Filter class that consists of several rules."""

    logical_functions = ["or", "and", "xor", "one"]

    def __init__(self, source=None):
        if source:
            self.need_param = source.need_param
            self.flist = source.flist[:]
            self.name = source.name
            self.comment = source.comment
            self.logical_op = source.logical_op
            self.invert = source.invert
        else:
            self.need_param = 0
            self.flist = []
            self.name = ""
            self.comment = ""
            self.logical_op = "and"
            self.invert = False

    def match(self, handle, db):
        """
        Return True or False depending on whether the handle matches the filter.
        """
        if self.apply(db, [handle]):
            return True
        else:
            return False

    def is_empty(self):
        return (len(self.flist) == 0) or (
            len(self.flist) == 1 and ((self.flist[0].is_empty() and not self.invert))
        )

    def set_logical_op(self, val):
        if val in GenericFilter.logical_functions:
            self.logical_op = val
        else:
            self.logical_op = "and"

    def get_logical_op(self):
        return self.logical_op

    def set_invert(self, val):
        self.invert = bool(val)

    def get_invert(self):
        return self.invert

    def get_name(self, ulocale=glocale):
        return self.name

    def set_name(self, name):
        self.name = name

    def set_comment(self, comment):
        self.comment = comment

    def get_comment(self):
        return self.comment

    def add_rule(self, rule):
        self.flist.append(rule)

    def delete_rule(self, rule):
        self.flist.remove(rule)

    def set_rules(self, rules):
        self.flist = rules

    def get_rules(self):
        return self.flist

    def get_cursor(self, db):
        return db.get_person_cursor()

    def get_tree_cursor(self, db):
        return db.get_person_cursor()

    def make_obj(self):
        return Person()

    def find_from_handle(self, db, handle):
        return db.get_person_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_people()

    def check_func(self, db, id_list, task, user=None, tupleind=None, tree=False):
        final_list = []
        if user:
            user.begin_progress(_("Filter"), _("Applying ..."), self.get_number(db))
        if id_list is None:
            with self.get_tree_cursor(db) if tree else self.get_cursor(db) as cursor:
                for handle, data in cursor:
                    person = self.make_obj()
                    person.unserialize(data)
                    if user:
                        user.step_progress()
                    if task(db, person) != self.invert:
                        final_list.append(handle)
        else:
            for data in id_list:
                if tupleind is None:
                    handle = data
                else:
                    handle = data[tupleind]
                person = self.find_from_handle(db, handle)
                if user:
                    user.step_progress()
                if task(db, person) != self.invert:
                    final_list.append(data)
        if user:
            user.end_progress()
        return final_list

    def check_and(self, db, id_list, user=None, tupleind=None, tree=False):
        final_list = []
        flist = self.flist
        if user:
            user.begin_progress(_("Filter"), _("Applying ..."), self.get_number(db))
        if id_list is None:
            with self.get_tree_cursor(db) if tree else self.get_cursor(db) as cursor:
                for handle, data in cursor:
                    person = self.make_obj()
                    person.unserialize(data)
                    if user:
                        user.step_progress()
                    val = all(rule.apply(db, person) for rule in flist)
                    if val != self.invert:
                        final_list.append(handle)
        else:
            for data in id_list:
                if tupleind is None:
                    handle = data
                else:
                    handle = data[tupleind]
                person = self.find_from_handle(db, handle)
                if user:
                    user.step_progress()
                val = all(rule.apply(db, person) for rule in flist if person)
                if val != self.invert:
                    final_list.append(data)
        if user:
            user.end_progress()
        return final_list

    def check_or(self, db, id_list, user=None, tupleind=None, tree=False):
        return self.check_func(db, id_list, self.or_test, user, tupleind, tree=False)

    def check_one(self, db, id_list, user=None, tupleind=None, tree=False):
        return self.check_func(db, id_list, self.one_test, user, tupleind, tree=False)

    def check_xor(self, db, id_list, user=None, tupleind=None, tree=False):
        return self.check_func(db, id_list, self.xor_test, user, tupleind, tree=False)

    def xor_test(self, db, person):
        test = False
        for rule in self.flist:
            test = test ^ rule.apply(db, person)
        return test

    def one_test(self, db, person):
        found_one = False
        for rule in self.flist:
            if rule.apply(db, person):
                if found_one:
                    return False  # There can be only one!
                found_one = True
        return found_one

    def or_test(self, db, person):
        return any(rule.apply(db, person) for rule in self.flist)

    def get_check_func(self):
        try:
            m = getattr(self, "check_" + self.logical_op)
        except AttributeError:
            m = self.check_and
        return m

    def check(self, db, handle):
        return self.get_check_func()(db, [handle])

    def apply(self, db, id_list=None, tupleind=None, user=None, tree=False):
        """
        Apply the filter using db.
        If id_list given, the handles in id_list are used. If not given
        a database cursor will be used over all entries.

        If tupleind is given, id_list is supposed to consist of a list of
        tuples, with the handle being index tupleind. So
        handle_0 = id_list[0][tupleind]

        user is optional. If present it must be an instance of a User class.

        :Returns: if id_list given, it is returned with the items that
                do not match the filter, filtered out.
                if id_list not given, all items in the database that
                match the filter are returned as a list of handles
        """
        m = self.get_check_func()
        for rule in self.flist:
            rule.requestprepare(db, user)
        res = m(db, id_list, user, tupleind, tree)
        for rule in self.flist:
            rule.requestreset()
        return res


class GenericFamilyFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_family_cursor()

    def make_obj(self):
        return Family()

    def find_from_handle(self, db, handle):
        return db.get_family_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_families()


class GenericEventFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_event_cursor()

    def make_obj(self):
        return Event()

    def find_from_handle(self, db, handle):
        return db.get_event_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_events()


class GenericSourceFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_source_cursor()

    def make_obj(self):
        return Source()

    def find_from_handle(self, db, handle):
        return db.get_source_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_sources()


class GenericCitationFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_citation_cursor()

    def get_tree_cursor(self, db):
        return db.get_citation_cursor()

    def make_obj(self):
        return Citation()

    def find_from_handle(self, db, handle):
        return db.get_citation_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_citations()


class GenericPlaceFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_place_cursor()

    def get_tree_cursor(self, db):
        return db.get_place_tree_cursor()

    def make_obj(self):
        return Place()

    def find_from_handle(self, db, handle):
        return db.get_place_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_places()


class GenericMediaFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_media_cursor()

    def make_obj(self):
        return Media()

    def find_from_handle(self, db, handle):
        return db.get_media_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_media()


class GenericRepoFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_repository_cursor()

    def make_obj(self):
        return Repository()

    def find_from_handle(self, db, handle):
        return db.get_repository_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_repositories()


class GenericNoteFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_note_cursor()

    def make_obj(self):
        return Note()

    def find_from_handle(self, db, handle):
        return db.get_note_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_notes()


def GenericFilterFactory(namespace):
    if namespace == "Person":
        return GenericFilter
    elif namespace == "Family":
        return GenericFamilyFilter
    elif namespace == "Event":
        return GenericEventFilter
    elif namespace == "Source":
        return GenericSourceFilter
    elif namespace == "Citation":
        return GenericCitationFilter
    elif namespace == "Place":
        return GenericPlaceFilter
    elif namespace == "Media":
        return GenericMediaFilter
    elif namespace == "Repository":
        return GenericRepoFilter
    elif namespace == "Note":
        return GenericNoteFilter


class DeferredFilter(GenericFilter):
    """
    Filter class allowing for deferred translation of the filter name
    """

    def __init__(self, filter_name, person_name):
        GenericFilter.__init__(self, None)
        self.name_pair = [filter_name, person_name]

    def get_name(self, ulocale=glocale):
        """
        return the filter name, possibly translated

        If ulocale is passed in (a :class:`.GrampsLocale`) then
        the translated value will be returned instead.

        :param ulocale: allow deferred translation of strings
        :type ulocale: a :class:`.GrampsLocale` instance
        """
        self._ = ulocale.translation.gettext
        if self.name_pair[1]:
            return self._(self.name_pair[0]) % self.name_pair[1]
        return self._(self.name_pair[0])


class DeferredFamilyFilter(GenericFamilyFilter):
    """
    Filter class allowing for deferred translation of the filter name
    """

    def __init__(self, filter_name, family_name):
        GenericFamilyFilter.__init__(self, None)
        self.name_pair = [filter_name, family_name]

    def get_name(self, ulocale=glocale):
        """
        return the filter name, possibly translated

        If ulocale is passed in (a :class:`.GrampsLocale`) then
        the translated value will be returned instead.

        :param ulocale: allow deferred translation of strings
        :type ulocale: a :class:`.GrampsLocale` instance
        """
        self._ = ulocale.translation.gettext
        if self.name_pair[1]:
            return self._(self.name_pair[0]) % self.name_pair[1]
        return self._(self.name_pair[0])



# Gramps - a GTK+/GNOME based genealogy program
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2009       Peter Landgren
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# plugins/mapservices/eniroswedenmap.py

"""
Eniro Sweden (Denmark) map service plugin. Opens place in kartor.eniro.se
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.plugins.lib.libmapservice import MapService
from gramps.gui.dialog import WarningDialog
from gramps.gen.utils.location import get_main_location
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.lib import PlaceType

# Make upper case of translated country so string search works later
MAP_NAMES_SWEDEN = [
    _("Sweden").upper(),
    "SVERIGE",
    "SWEDEN",
    "SUEDOIS",
    "ROUTSI",
    "SCHWEDEN",
]
MAP_NAMES_DENMARK = [
    _("Denmark").upper(),
    "DANMARK",
    "DENMARK",
    "DANOIS",
    "TANSKA",
    "DÄNEMARK",
]


def _strip_leading_comma(descr):
    """Strips leading comma
    and leading and trailing spaces
    """
    if len(descr) > 0 and descr.strip()[0] == ",":
        descr = descr.strip()[1:]
    return descr.strip()


def _build_title(db, place):
    """Builds descrition string for title parameter in url"""
    descr = place_displayer.display(db, place)
    location = get_main_location(db, place)
    parish = location.get(PlaceType.PARISH)
    city = location.get(PlaceType.CITY)
    state = location.get(PlaceType.STATE)
    title_descr = ""
    if descr:
        title_descr += descr.strip()
    if parish:
        # TODO for Arabic, should the next line's comma be translated?
        title_descr += ", " + parish.strip() + _(" parish")
    if city:
        # TODO for Arabic, should the next line's comma be translated?
        title_descr += ", " + city.strip()
    if state:
        # TODO for Arabic, should the next line's comma be translated?
        title_descr += ", " + state.strip() + _(" state")
    return _strip_leading_comma(title_descr)


def _build_city(db, place):
    """Builds description string for city parameter in url"""
    location = get_main_location(db, place)
    county = location.get(PlaceType.COUNTY)
    # Build a title description string that will work for Eniro
    city_descr = _build_area(db, place)
    if county:
        # TODO for Arabic, should the next line's comma be translated?
        city_descr += ", " + county
    return _strip_leading_comma(city_descr)


def _build_area(db, place):
    """Builds string for area parameter in url"""
    location = get_main_location(db, place)
    street = location.get(PlaceType.STREET)
    city = location.get(PlaceType.CITY)
    # Build a title description string that will work for Eniro
    area_descr = ""
    if street:
        area_descr += street.strip()
    if city:
        # TODO for Arabic, should the next line's comma be translated?
        area_descr += ", " + city
    return _strip_leading_comma(area_descr)


class EniroSVMapService(MapService):
    """Map  service using http://kartor.eniro.se"""

    def __init__(self):
        MapService.__init__(self)

    def calc_url(self):
        """Determine the url to use on maps.google.com
        Logic: valid for places within Sweden and
               Denmark, only if lat lon avalible
               use lat lon if present
               otherwise use city and country if present
               otherwise use description of the place
        """
        place = self._get_first_place()[0]
        path = ""
        # First see if we are in or near Sweden or Denmark
        # Change country to upper case
        location = get_main_location(self.database, place)
        country = location.get(PlaceType.COUNTRY, "").upper().strip()
        country_given = (
            country in MAP_NAMES_SWEDEN or country in MAP_NAMES_DENMARK
        ) and (country != "")
        # if no country given, check if we might be in the vicinity defined by
        # 54 33' 0" < lat < 66 9' 0", 54.55 and 69.05
        # 8 3' 0" < long < 24 9' 0", 8.05 and 24.15
        latitude, longitude = self._lat_lon(place)
        if latitude is None or longitude is None:
            coord_ok = False
        else:
            latitude = float(latitude)
            longitude = float(longitude)
            # Check if coordinates are inside Sweden and Denmark
            if (54.55 < latitude < 69.05) and (8.05 < longitude < 24.15):
                coord_ok = True
            else:
                msg2 = _("Latitude not within '54.55' to '69.05'\n") + _(
                    "Longitude not within '8.05' to '24.15'"
                )
                WarningDialog(
                    _("Eniro map not available"), msg2, parent=self.uistate.window
                )
                return

        if coord_ok:
            place_title = _build_title(self.database, place)
            place_city = _build_city(self.database, place)
            x_coord, y_coord = self._lat_lon(place, format="RT90")
            # Set zoom level to 5 if Sweden/Denmark, others 3
            zoom = 5
            if not country_given:
                zoom = 3
            path = (
                "http://www.eniro.se/partner.fcgi?pis=1&x=%s&y=%s"
                "&zoom_level=%i&map_size=0&title=%s&city=%s&partner=gramps"
            )
            # Note x and y are swapped!
            path = path % (y_coord, x_coord, zoom, place_title, place_city)
            self.url = path.replace(" ", "%20")
            return

        place_area = _build_area(self.database, place)
        if country_given and place_area:
            if country in MAP_NAMES_SWEDEN:
                path = (
                    "http://kartor.eniro.se/query?&what=map_adr&mop=aq"
                    "&geo_area=%s&partner=gramps"
                )
                path = path % (place_area)
                self.url = path.replace(" ", "%20")
                return
            else:
                WarningDialog(
                    _("Eniro map not available"),
                    _("Coordinates needed in Denmark"),
                    parent=self.uistate.window,
                )
                self.url = ""
                return

        WarningDialog(
            _("Eniro map not available"),
            _("Latitude and longitude,\n" "or street and city needed"),
            parent=self.uistate.window,
        )
        return



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from .editsecondary import EditSecondary
from ..glade import Glade
from ..widgets import MonitoredEntry
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# EditLocation class
#
# -------------------------------------------------------------------------
class EditLocation(EditSecondary):
    def __init__(self, dbstate, uistate, track, location, callback):
        EditSecondary.__init__(self, dbstate, uistate, track, location, callback)

    def _local_init(self):
        self.top = Glade()
        self.set_window(self.top.toplevel, None, _("Location Editor"))
        self.setup_configs("interface.location", 600, 250)

    def _setup_fields(self):
        self.street = MonitoredEntry(
            self.top.get_object("street"),
            self.obj.set_street,
            self.obj.get_street,
            self.db.readonly,
        )

        self.locality = MonitoredEntry(
            self.top.get_object("locality"),
            self.obj.set_locality,
            self.obj.get_locality,
            self.db.readonly,
        )

        self.city = MonitoredEntry(
            self.top.get_object("city"),
            self.obj.set_city,
            self.obj.get_city,
            self.db.readonly,
        )

        self.state = MonitoredEntry(
            self.top.get_object("state"),
            self.obj.set_state,
            self.obj.get_state,
            self.db.readonly,
        )

        self.postal = MonitoredEntry(
            self.top.get_object("postal"),
            self.obj.set_postal_code,
            self.obj.get_postal_code,
            self.db.readonly,
        )

        self.phone = MonitoredEntry(
            self.top.get_object("phone"),
            self.obj.set_phone,
            self.obj.get_phone,
            self.db.readonly,
        )

        self.parish = MonitoredEntry(
            self.top.get_object("parish"),
            self.obj.set_parish,
            self.obj.get_parish,
            self.db.readonly,
        )

        self.county = MonitoredEntry(
            self.top.get_object("county"),
            self.obj.set_county,
            self.obj.get_county,
            self.db.readonly,
        )

        self.country = MonitoredEntry(
            self.top.get_object("country"),
            self.obj.set_country,
            self.obj.get_country,
            self.db.readonly,
        )

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("button119"))
        self.define_ok_button(self.top.get_object("button118"), self.save)
        self.define_help_button(self.top.get_object("button128"))

    def save(self, *obj):
        if self.callback:
            self.callback(self.obj)
        self.close()



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-now       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
import logging
import xml.etree.ElementTree as ETree
from gi.repository import GObject

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
_LOG = logging.getLogger("maps.libkml")


class Kml(GObject.GObject):
    """
    This is the library used to read kml files
    * Allowed : points, paths and polygons.

    * One point : name, (latitude, longitude)
    * One path  : name, [ (latitude, longitude), (latitude, longitude), ...]
    * One polygon : name, type, color,
    *               [ (latitude, longitude), (latitude, longitude), ...]

    * Some kml files use the altitude. We don't use it.

    """

    def __init__(self, kml_file):
        """
        Initialize the library
        The access right and validity of the kmlfile must be verified
        before this method.
        """
        GObject.GObject.__init__(self)
        self.tag = ""
        self.name = ""
        self.color = ""
        self.transparency = ""
        self.type = ""
        self.points = []
        self.markers = []
        self.paths = []
        self.polygons = []
        self.tree = ETree.parse(kml_file)
        root = self.tree.getroot()
        self.tag = root.tag.replace("}kml", "}")
        _LOG.debug("Tag version of kml file %s is %s", kml_file, self.tag)
        fname, dummy_extension = os.path.splitext(kml_file)
        dummy_fdir, self.kmlfile = os.path.split(fname)

    def set_default(self):
        """
        Set all the default value for the marker
        """
        self.name = self.kmlfile
        self.type = None
        self.color = None
        self.transparency = None

    def get_color(self, attributes):
        """
        Get the color for this marker
        """
        pass

    def get_transparency(self, attributes):
        """
        Get the transparency for this marker
        """
        pass

    def get_coordinates(self, attributes):
        """
        Get all the coordinates for this marker
        """
        self.points = []
        for point in attributes.text.split():
            try:
                (longitude, latitude, dummy_altitude) = point.split(",")
            except:
                (longitude, latitude) = point.split(",")
            self.points.append((float(latitude), float(longitude)))

    def get_linear_ring(self, attributes):
        """
        Get all the coordinates for this marker
        """
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "coordinates":
                self.get_coordinates(sub_attribute)

    def get_polygon_outer_boundary(self, attributes):
        """
        This function get the coordinates used to draw a filled polygon.
        """
        self.type = "OuterPolygon"
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "LinearRing":
                self.get_linear_ring(sub_attribute)

    def get_polygon_inner_boundary(self, attributes):
        """
        This function get the coordinates used to draw a hole inside
        a filled polygon.
        """
        self.type = "InnerPolygon"
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "LinearRing":
                self.get_linear_ring(sub_attribute)

    def get_polygon(self, attributes):
        """
        Get all the coordinates for the polygon
        """
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "outerBoundaryIs":
                self.get_polygon_outer_boundary(sub_attribute)
                self.polygons.append(
                    (self.name, self.type, self.color, self.transparency, self.points)
                )
            if sub_attribute.tag == self.tag + "innerBoundaryIs":
                self.get_polygon_inner_boundary(sub_attribute)
                self.polygons.append(
                    (self.name, self.type, self.color, self.transparency, self.points)
                )
            if sub_attribute.tag == self.tag + "LinearRing":
                self.get_linear_ring(sub_attribute)
                self.type = "Polygon"
                self.polygons.append(
                    (self.name, self.type, self.color, self.transparency, self.points)
                )

    def get_point(self, attributes):
        """
        Get all the coordinates for this marker
        """
        self.type = "Point"
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "coordinates":
                self.get_coordinates(sub_attribute)

    def get_path(self, attributes):
        """
        Get all the coordinates for this marker
        """
        self.type = "Path"
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "coordinates":
                self.get_coordinates(sub_attribute)
                self.paths.append(
                    (self.name, self.type, self.color, self.transparency, self.points)
                )

    def get_multi_geometry(self, attributes):
        """
        Do we have some sub structures ?
        """
        for sub_attribute in attributes:
            if sub_attribute.tag == self.tag + "Polygon":
                self.get_polygon(sub_attribute)
            if sub_attribute.tag == self.tag + "LineString":
                self.get_path(sub_attribute)

    def add_kml(self):
        """
        Add a kml file.
        """
        line_strings = self.tree.findall(".//" + self.tag + "Placemark")
        for attributes in line_strings:
            self.points = []
            self.set_default()
            for sub_attribute in attributes:
                if sub_attribute.tag == self.tag + "name":
                    self.name = sub_attribute.text
                if sub_attribute.tag == self.tag + "Polygon":
                    self.get_polygon(sub_attribute)
                if sub_attribute.tag == self.tag + "LineString":
                    self.get_path(sub_attribute)
                if sub_attribute.tag == self.tag + "MultiGeometry":
                    self.get_multi_geometry(sub_attribute)
        return (self.paths, self.polygons)

    def add_points(self):
        """
        Add points or markers
        """
        line_strings = self.tree.findall(".//" + self.tag + "Placemark")
        self.markers = []
        for attributes in line_strings:
            self.points = []
            self.set_default()
            for sub_attribute in attributes:
                if sub_attribute.tag == self.tag + "name":
                    self.name = sub_attribute.text
                if sub_attribute.tag == self.tag + "Point":
                    self.get_point(sub_attribute)
                    self.markers.append((self.name, self.points))
        return self.markers


GObject.type_register(Kml)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import pickle

import logging

_LOG = logging.getLogger(".objectentries")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Person, Place, Source, Media, Note
from .editperson import EditPerson
from .editplace import EditPlace
from .editsource import EditSource
from .editmedia import EditMedia
from .editnote import EditNote
from ..selectors import SelectorFactory
from ..ddtargets import DdTargets
from gramps.gen.errors import WindowActiveError
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.display.name import displayer as name_displayer


# -------------------------------------------------------------------------
#
# ObjEntry
#
# -------------------------------------------------------------------------
class ObjEntry:
    """
    Handles the selection of a existing or new Object. Supports Drag and Drop
    to select the object.
    This is the base class to create a real entry
    """

    EMPTY_TEXT = ""
    EMPTY_TEXT_RED = ""
    EDIT_STR = ""
    SHARE_STR = ""
    ADD_STR = ""
    DEL_STR = ""
    _DND_TYPE = None
    _DND_ICON = None

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        label,
        label_event_box,
        set_val,
        get_val,
        add_edt,
        share,
        callback=None,
    ):
        """Pass the dbstate and uistate and present track.
        label is a Gtk.Label that shows the persent value
        label_event_box is a Gtk.EventBox used for allowing drag from the label
        set_val is function that is called when handle changes, use it
            to update the calling module
        get_val is function that is called to obtain handle from calling
            module
        share is the Gtk.Button to call the object selector or del connect
        add_edt is the Gtk.Button with add or edit value. Pass None if
            this button should not be present.
        """
        self.label = label
        self.label_event_box = label_event_box
        self.add_edt = add_edt
        self.share = share
        self.dbstate = dbstate
        self.db = dbstate.db
        self.get_val = get_val
        self.set_val = set_val
        self.uistate = uistate
        self.track = track
        self.callback = callback

        # set the object specific code
        self._init_object()

        # check if valid object:
        handle = self.get_val()
        if handle:
            obj = self.get_from_handle(handle)
            if not obj:
                # invalid val, set it to None
                self.set_val(None)
        if self.get_val():
            self.set_button(True)
            obj = self.get_from_handle(self.get_val())
            name = self.get_label(obj)
        else:
            name = ""
            self.set_button(False)

        if self.db.readonly:
            if self.add_edt is not None:
                self.add_edt.set_sensitive(False)
            self.share.set_sensitive(False)
        else:
            if self.add_edt is not None:
                self.add_edt.set_sensitive(True)
            self.share.set_sensitive(True)

        if self.add_edt is not None:
            self.add_edt.connect("clicked", self.add_edt_clicked)
        self.share.connect("clicked", self.share_clicked)

        if not self.db.readonly and not name:
            if self.add_edt is None:
                self.label.set_text(self.EMPTY_TEXT_RED)
            else:
                self.label.set_text(self.EMPTY_TEXT)
            self.label.set_use_markup(True)
        else:
            self.label.set_text(name)
        self.label.set_ellipsize(Pango.EllipsizeMode.END)
        if self.callback:
            self.callback()

    def _update_dnd_capability(self):
        """
        Connect drag and drop of object entry
            Drop is always allowed
            Drag only allowed when object is set
        """
        if not self._DND_TYPE:
            return
        if not self.label.drag_dest_get_target_list():
            self.label.drag_dest_set(
                Gtk.DestDefaults.ALL, [self._DND_TYPE.target()], Gdk.DragAction.COPY
            )
            self.label.connect("drag_data_received", self.drag_data_received)
        # Set the drag action from this box
        if self.get_val():
            if not self.label.drag_source_get_target_list():
                self.label_event_box.drag_source_set(
                    Gdk.ModifierType.BUTTON1_MASK,
                    [self._DND_TYPE.target()],
                    Gdk.DragAction.COPY,
                )
                if self._DND_ICON:
                    self.label_event_box.drag_source_set_icon_name(self._DND_ICON)
                self.label_event_box.connect("drag_data_get", self.drag_data_get)
        else:
            if self.label.drag_source_get_target_list():
                self.label_event_box.drag_source_unset()

    def _init_object(self):
        """inheriting objects can use this to set extra variables"""
        pass

    def get_from_handle(self, handle):
        """return the object given the hande
        inheriting objects must set this
        """
        pass

    def get_label(self, object):
        """return the label
        inheriting objects must set this
        """
        pass

    def after_edit(self, obj):
        name = self.get_label(obj)
        self.label.set_text(name)
        if self.callback:
            self.callback()

    def add_edt_clicked(self, obj):
        """if value, edit, if no value, call editor on new object"""
        if self.get_val():
            obj = self.get_from_handle(self.get_val())
            self.call_editor(obj)
        else:
            self.call_editor()

    def call_editor(self, obj):
        """inheriting objects must set this"""
        pass

    def call_selector(self):
        """inheriting objects must set this"""
        pass

    def drag_data_received(self, widget, context, x, y, selection, info, time):
        (drag_type, idval, obj, val) = pickle.loads(selection.get_data())
        data = self.get_from_handle(obj)
        self.obj_added(data)

    def drag_data_get(self, widget, context, sel_data, info, time):
        # get the selected object, returning if not is defined
        if info == self._DND_TYPE.app_id:
            data = (self._DND_TYPE.drag_type, id(self), self.get_val(), 0)
            sel_data.set(self._DND_TYPE.atom_drag_type, 8, pickle.dumps(data))

    def obj_added(self, data):
        """callback from adding an object to the entry"""
        self.set_val(data.handle)
        self.label.set_text(self.get_label(data))
        self.set_button(True)
        if self.callback:
            self.callback()

    def share_clicked(self, obj):
        """if value, delete connect, in no value, select existing object"""
        if self.get_val():
            self.set_val(None)
            self.label.set_text(self.EMPTY_TEXT)
            self.label.set_use_markup(True)
            self.set_button(False)
            if self.callback:
                self.callback()
        else:
            select = self.call_selector()
            obj = select.run()
            if obj:
                self.obj_added(obj)

    def set_button(self, use_add):
        """This sets the correct image to the two buttons.
        If False: select icon and add icon
        If True:  remove icon and edit icon
        """
        if self.add_edt is not None:
            list(map(self.add_edt.remove, self.add_edt.get_children()))
        for i in self.share.get_children():
            self.share.remove(i)

        if use_add:
            image = Gtk.Image()
            image.set_from_icon_name("list-remove", Gtk.IconSize.BUTTON)
            image.show()
            self.share.add(image)
            self.share.set_tooltip_text(self.DEL_STR)
            if self.add_edt is not None:
                image = Gtk.Image()
                image.set_from_icon_name("gtk-edit", Gtk.IconSize.BUTTON)
                image.show()
                self.add_edt.add(image)
                self.add_edt.set_tooltip_text(self.EDIT_STR)
        else:
            image = Gtk.Image()
            image.set_from_icon_name("gtk-index", Gtk.IconSize.BUTTON)
            image.show()
            self.share.add(image)
            self.share.set_tooltip_text(self.SHARE_STR)
            if self.add_edt is not None:
                image = Gtk.Image()
                image.set_from_icon_name("list-add", Gtk.IconSize.BUTTON)
                image.show()
                self.add_edt.add(image)
                self.add_edt.set_tooltip_text(self.ADD_STR)
        self._update_dnd_capability()


class PersonEntry(ObjEntry):
    """
    Handles the selection of a existing or new Person. Supports Drag and Drop
    to select a person.
    """

    EMPTY_TEXT = "<i>%s</i>" % _(
        "To select a person, use drag-and-drop " "or use the buttons"
    )
    EMPTY_TEXT_RED = "<i>%s</i>" % _("No person given, click button to select one")
    EDIT_STR = _("Edit person")
    SHARE_STR = _("Select an existing person")
    ADD_STR = _("Add a new person")
    DEL_STR = _("Remove person")
    _DND_TYPE = DdTargets.PERSON_LINK
    _DND_ICON = "gramps-person"

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        label,
        label_event_box,
        set_val,
        get_val,
        add_edt,
        share,
        skip=[],
    ):
        ObjEntry.__init__(
            self,
            dbstate,
            uistate,
            track,
            label,
            label_event_box,
            set_val,
            get_val,
            add_edt,
            share,
        )
        self.skip = skip

    def get_from_handle(self, handle):
        """return the object given the hande"""
        return self.db.get_person_from_handle(handle)

    def get_label(self, person):
        person_name = name_displayer.display(person)
        return "%s \u2068[%s]\u200e\u2069" % (person_name, person.gramps_id)

    def call_editor(self, obj=None):
        if obj is None:
            person = Person()
            func = self.obj_added
        else:
            person = obj
            func = self.after_edit
        try:
            EditPerson(self.dbstate, self.uistate, self.track, person, func)
        except WindowActiveError:
            pass

    def call_selector(self):
        cls = SelectorFactory("Person")
        return cls(self.dbstate, self.uistate, self.track, skip=self.skip)


class PlaceEntry(ObjEntry):
    """
    Handles the selection of a existing or new Place. Supports Drag and Drop
    to select a place.
    """

    EMPTY_TEXT = "<i>%s</i>" % _(
        "To select a place, use drag-and-drop " "or use the buttons"
    )
    EMPTY_TEXT_RED = "<i>%s</i>" % _("No place given, click button to select one")
    EDIT_STR = _("Edit place")
    SHARE_STR = _("Select an existing place")
    ADD_STR = _("Add a new place")
    DEL_STR = _("Remove place")
    _DND_TYPE = DdTargets.PLACE_LINK
    _DND_ICON = "gramps-place"

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        label,
        label_event_box,
        set_val,
        get_val,
        add_edt,
        share,
        skip=[],
    ):
        ObjEntry.__init__(
            self,
            dbstate,
            uistate,
            track,
            label,
            label_event_box,
            set_val,
            get_val,
            add_edt,
            share,
        )
        self.skip = skip

    def get_from_handle(self, handle):
        """return the object given the hande"""
        return self.db.get_place_from_handle(handle)

    def get_label(self, place):
        place_title = place_displayer.display(self.db, place)
        # When part of the place title contains RTL text, the gramps_id gets
        # messed up; so use Unicode FSI/PDI and LRM chars to isolate it.
        # see bug 10124 and PR924
        return "%s \u2068[%s]\u200e\u2069" % (place_title, place.gramps_id)

    def call_editor(self, obj=None):
        if obj is None:
            place = Place()
            func = self.obj_added
        else:
            place = obj
            func = self.after_edit
        try:
            EditPlace(self.dbstate, self.uistate, self.track, place, func)
        except WindowActiveError:
            pass

    def call_selector(self):
        cls = SelectorFactory("Place")
        return cls(self.dbstate, self.uistate, self.track, skip=self.skip)


class SourceEntry(ObjEntry):
    """
    Handles the selection of a existing or new Source. Supports Drag and Drop
    to select a source.
    """

    EMPTY_TEXT = "<i>%s</i>" % _(
        "To select a source, use drag-and-drop " "or use the buttons"
    )
    EMPTY_TEXT_RED = "<i>%s</i>" % _("First add a source using the button")
    EDIT_STR = _("Edit source")
    SHARE_STR = _("Select an existing source")
    ADD_STR = _("Add a new source")
    DEL_STR = _("Remove source")
    _DND_TYPE = DdTargets.SOURCE_LINK
    _DND_ICON = "gramps-source"

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        label,
        label_event_box,
        set_val,
        get_val,
        add_edt,
        share,
        callback,
    ):
        ObjEntry.__init__(
            self,
            dbstate,
            uistate,
            track,
            label,
            label_event_box,
            set_val,
            get_val,
            add_edt,
            share,
            callback,
        )

    def get_from_handle(self, handle):
        """return the object given the handle"""
        return self.db.get_source_from_handle(handle)

    def get_label(self, source):
        return "%s \u2068[%s]\u200e\u2069" % (source.get_title(), source.gramps_id)

    def call_editor(self, obj=None):
        if obj is None:
            source = Source()
            func = self.obj_added
        else:
            source = obj
            func = self.after_edit
        try:
            EditSource(self.dbstate, self.uistate, self.track, source, func)
        except WindowActiveError:
            pass

    def call_selector(self):
        cls = SelectorFactory("Source")
        return cls(self.dbstate, self.uistate, self.track)


# FIXME isn't used anywhere
class MediaEntry(ObjEntry):
    """
    Handles the selection of a existing or new media. Supports Drag and Drop
    to select a media object.
    """

    EMPTY_TEXT = "<i>%s</i>" % _(
        "To select a media object, use drag-and-drop " "or use the buttons"
    )
    EMPTY_TEXT_RED = "<i>%s</i>" % _("No image given, click button to select one")
    EDIT_STR = _("Edit media object")
    SHARE_STR = _("Select an existing media object")
    ADD_STR = _("Add a new media object")
    DEL_STR = _("Remove media object")
    _DND_TYPE = DdTargets.MEDIAOBJ
    _DND_ICON = "gramps-media"

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        label,
        label_event_box,
        set_val,
        get_val,
        add_edt,
        share,
    ):
        ObjEntry.__init__(
            self,
            dbstate,
            uistate,
            track,
            label,
            label_event_box,
            set_val,
            get_val,
            add_edt,
            share,
        )

    def get_from_handle(self, handle):
        """return the object given the hande"""
        return self.db.get_media_from_handle(handle)

    def get_label(self, object):
        return "%s \u2068[%s]\u200e\u2069" % (
            object.get_description(),
            object.gramps_id,
        )

    def call_editor(self, obj=None):
        if obj is None:
            object = Media()
            func = self.obj_added
        else:
            object = obj
            func = self.after_edit
        try:
            EditMedia(self.dbstate, self.uistate, self.track, object, func)
        except WindowActiveError:
            pass

    def call_selector(self):
        cls = SelectorFactory("Media")
        return cls(self.dbstate, self.uistate, self.track)


# FIXME isn't used anywhere
class NoteEntry(ObjEntry):
    """
    Handles the selection of a existing or new Note. Supports Drag and Drop
    to select a Note.
    """

    EMPTY_TEXT = "<i>%s</i>" % _(
        "To select a note, use drag-and-drop " "or use the buttons"
    )
    EMPTY_TEXT_RED = "<i>%s</i>" % _("No note given, click button to select one")
    EDIT_STR = _("Edit Note")
    SHARE_STR = _("Select an existing note")
    ADD_STR = _("Add a new note")
    DEL_STR = _("Remove note")
    _DND_TYPE = DdTargets.NOTE_LINK
    _DND_ICON = "gramps-notes"

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        label,
        label_event_box,
        set_val,
        get_val,
        add_edt,
        share,
    ):
        ObjEntry.__init__(
            self,
            dbstate,
            uistate,
            track,
            label,
            label_event_box,
            set_val,
            get_val,
            add_edt,
            share,
        )
        self.notetype = None

    def set_notetype(self, type):
        """set a notetype to use in new notes"""
        self.notetype = type

    def get_notetype(self):
        """return the set notetype"""
        return self.notetype

    def get_from_handle(self, handle):
        """return the object given the hande"""
        return self.db.get_note_from_handle(handle)

    def get_label(self, note):
        txt = " ".join(note.get().split())
        if len(txt) > 35:
            txt = txt[:35] + "..."
        return "%s \u2068[%s]\u200e\u2069" % (txt, note.gramps_id)

    def call_editor(self, obj=None):
        if obj is None:
            note = Note()
            note.set_type(self.get_notetype())
            func = self.obj_added
        else:
            note = obj
            func = self.after_edit
        try:
            EditNote(self.dbstate, self.uistate, self.track, note, func)
        except WindowActiveError:
            pass

    def call_selector(self):
        cls = SelectorFactory("Note")
        return cls(self.dbstate, self.uistate, self.track)
